from importlib import resources
from unittest.mock import patch

from astropy import table
from astropy.io import fits
from celery.exceptions import Ignore
import numpy as np
import pytest

from . import data
from ..tasks.bayestar import localize
from ..util.tempfile import NamedTemporaryFile


def mock_bayestar(event, *args, **kwargs):
    # Attempt to access single-detector triggers, so that a
    # DetectorDisabledError may be raised
    event.singles

    return table.Table({'UNIQ': np.arange(4, 16, dtype=np.int64),
                        'PROBDENSITY': np.ones(12),
                        'DISTMU': np.ones(12),
                        'DISTSIGMA': np.ones(12),
                        'DISTNORM': np.ones(12)})


@pytest.fixture
def coinc():
    return resources.read_binary(data, 'coinc.xml')


@patch('ligo.skymap.bayestar.localize', mock_bayestar)
def test_localize(coinc):
    """Test running BAYESTAR on G211117"""
    fitscontent = localize(coinc, 'G211117')
    with NamedTemporaryFile(content=fitscontent) as fitsfile:
        url = fits.getval(fitsfile.name, 'REFERENC', 1)
        assert url == 'https://gracedb.invalid/events/G211117'


@patch('ligo.skymap.bayestar.localize', mock_bayestar)
def test_localize_all_detectors_disabled(coinc):
    """Test running BAYESTAR on G211117, all detectors disabled"""
    with pytest.raises(Ignore):
        localize(coinc, 'G211117', disabled_detectors=['H1', 'L1', 'V1'])



"""VOEvent-related utilities."""
import ipaddress
import socket


def get_host_port(address):
    """Split a network address of the form ``host:port``.

    Parameters
    ----------
    network : str
        The network address.

    Returns
    -------
    host : str
        The hostname, or an empty string if missing.
    port : int, None
        The port number, or None if missing.

    """
    host, _, port = address.partition(':')
    return host, (int(port) if port else None)


def get_local_ivo(app):
    """Create an IVOID to identify this application in VOEvent Transport
    Protocol packets.

    Returns
    -------
    str
        A local IVOID composed of the machine's fully qualified domain name and
        the Celery application name (for example,
        `ivo://emfollow.ligo.caltech.edu/gwcelery`).

    """
    return 'ivo://{}/{}'.format(socket.getfqdn(), app.main)


def get_network(address):
    """Find the IP network prefix for a hostname or CIDR notation.

    Parameters
    ----------
    address : str
        A hostname, such as ``ligo.org``, or an IP address prefix in CIDR
        notation, such as ``127.0.0.0/8``.

    Returns
    -------
    ipaddress.IPv4Network
        An object representing the IP address prefix.

    """
    try:
        net = ipaddress.ip_network(address, strict=False)
    except ValueError:
        net = ipaddress.ip_network(socket.gethostbyname(address), strict=False)
    return net



from io import BytesIO
import logging
from pathlib import Path
from unittest.mock import call, patch

from astropy.time import Time
from gwpy.timeseries import Bits
import matplotlib.pyplot as plt
import numpy as np
import pytest

from ..import app
from ..import _version
from ..tasks import detchar


@pytest.fixture
def llhoft_glob_pass(monkeypatch):
    path = str(Path(__file__).parent / 'data/llhoft/pass/{detector}/*.gwf')
    yield monkeypatch.setitem(app.conf, 'llhoft_glob', path)


@pytest.fixture
def llhoft_glob_fail(monkeypatch):
    path = str(Path(__file__).parent / 'data/llhoft/fail/{detector}/*.gwf')
    yield monkeypatch.setitem(app.conf, 'llhoft_glob', path)


@pytest.fixture
def ifo_h1(monkeypatch):
    monkeypatch.setitem(app.conf, 'llhoft_channels', {
        'H1:DMT-DQ_VECTOR': 'dmt_dq_vector_bits',
        'H1:GDS-CALIB_STATE_VECTOR': 'ligo_state_vector_bits'})


@pytest.fixture
def ifo_h1_idq(monkeypatch):
    monkeypatch.setitem(
        app.conf, 'idq_channels', ['H1:IDQ-FAP_OVL_32_2048'])


@pytest.fixture
def gatedpipe(monkeypatch):
    monkeypatch.setitem(app.conf, 'uses_gatedhoft', {'gatepipe': True})


@pytest.fixture
def gatedpipe_prepost(monkeypatch):
    monkeypatch.setitem(
        app.conf, 'check_vector_prepost', {'gatepipe': [0.5, 0.5]})


@pytest.fixture
def scan_strainname(monkeypatch):
    monkeypatch.setitem(
        app.conf, 'strain_channel_names', {'H1': 'H1:GWOSC-16KHZ_R1_STRAIN'})


def test_create_cache(llhoft_glob_fail):
    assert len(detchar.create_cache('L1', 1216577976, 1216577979)) == 1


@patch('gwcelery.tasks.detchar.find_urls', return_value=[])
def test_create_cache_old_data(mock_find, llhoft_glob_fail):
    start, end = 1198800018, 1198800028
    detchar.create_cache('L1', start, end)
    mock_find.assert_called()


expected_path = Path(__file__).parent / 'data/llhoft/omegascan/scanme.gwf'


@patch('gwcelery.tasks.detchar.create_cache', return_value=[expected_path])
def test_make_omegascan_worked(mock_create_cache, scan_strainname):
    durs = [(1, 1), (1, 1), (1, 1)]
    t0 = 1126259463
    png = detchar.make_omegascan('H1', t0, durs)
    pngarray = plt.imread(BytesIO(png))
    # Test to see that the png is wider than 2000 pixels, indicating
    # presence of omegascan(s)
    assert plt.imshow(pngarray).get_extent()[1] > 850


@patch('gwcelery.tasks.detchar.create_cache', return_value=[])
def test_make_omegascan_failed(mock_create_cache, scan_strainname):
    durs = [(1, 1), (1, 1), (1, 1)]
    t0 = 1126259463
    png = detchar.make_omegascan('H1', t0, durs)
    pngarray = plt.imread(BytesIO(png))
    # Test to see that the png is narrower than 2000 pixels
    assert plt.imshow(pngarray).get_extent()[1] < 2000


@patch('gwcelery.tasks.detchar.make_omegascan.run',
       return_value=BytesIO().getvalue())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_omegascan(mock_upload, mock_fig):
    t0 = 1126259463
    gid = "S1234"
    detchar.omegascan(t0, gid)
    mock_upload.assert_called_with(
        mock_fig(), "V1_omegascan.png", gid, "V1 omegascan",
        tags=['data_quality']
    )


def test_omegascan_skips_ew(caplog):
    """Test that omegascans are delayed for events in the future."""
    caplog.set_level(logging.INFO)
    detchar.omegascan(Time.now().gps + 99, 'S1234')
    messages = [record.message for record in caplog.records]
    assert 'Delaying omegascan because S1234 is in the future' in messages  # noqa: E501


def test_check_idq(llhoft_glob_pass):
    channel = 'H1:IDQ-FAP_OVL_32_2048'
    start, end = 1216577976, 1216577980
    cache = detchar.create_cache('H1', start, end)
    assert detchar.check_idq(cache, channel, start, end) == (
        'H1:IDQ-FAP_OVL_32_2048', 1)


@patch('time.strftime', return_value='00:00:00 UTC Mon 01 Jan 2000')
@patch('socket.gethostname', return_value='test_host')
@patch('getpass.getuser', return_value='test_user')
def test_dqr_json(mock_time, mock_host, mock_user):
    state = "pass"
    summary = "72 and sunny!!"
    assert detchar.dqr_json(state, summary) == {
        'state': 'pass',
        'process_name': 'gwcelery.tasks.detchar',
        'process_version': _version.get_versions()['version'],
        'librarian': 'Geoffrey Mo (geoffrey.mo@ligo.org)',
        'date': '00:00:00 UTC Mon 01 Jan 2000',
        'hostname': 'test_host',
        'username': 'test_user',
        'summary': '72 and sunny!!',
        'figures': [],
        'tables': [],
        'links': ([
            {
                "href":
                "https://gwcelery.readthedocs.io/en/latest/gwcelery.tasks.detchar.html#gwcelery.tasks.detchar.check_vectors", # noqa
                "innerHTML": "a link to the documentation for this process"
            },
            {
                "href":
                "https://git.ligo.org/emfollow/gwcelery/blob/main/gwcelery/tasks/detchar.py",  # noqa
                "innerHTML": "a link to the source code in the gwcelery repo"
            }
        ]),
        'extra': []
    }


def test_check_vector(llhoft_glob_pass):
    channel = 'H1:DMT-DQ_VECTOR'
    start, end = 1216577976, 1216577980
    cache = detchar.create_cache('H1', start, end)
    bit_defs = {channel_type: Bits(channel=bitdef['channel'],
                                   bits=bitdef['bits'])
                for channel_type, bitdef
                in app.conf['detchar_bit_definitions'].items()}
    assert detchar.check_vector(cache, channel, start, end,
                                bit_defs['dmt_dq_vector_bits']) == {
        'H1:NO_OMC_DCPD_ADC_OVERFLOW': True,
        'H1:NO_DMT-ETMY_ESD_DAC_OVERFLOW': True}


@patch('gwcelery.tasks.detchar.StateVector.read', return_value=np.asarray([]))
def test_check_vector_fails_on_empty(mock_statevector, llhoft_glob_pass):
    channel = 'H1:DMT-DQ_VECTOR'
    start, end = 1216577976, 1216577980
    cache = detchar.create_cache('H1', start, end)
    bit_defs = {channel_type: Bits(channel=bitdef['channel'],
                                   bits=bitdef['bits'])
                for channel_type, bitdef
                in app.conf['detchar_bit_definitions'].items()}
    assert detchar.check_vector(cache, channel, start, end,
                                bit_defs['dmt_dq_vector_bits']) == {
        'H1:NO_OMC_DCPD_ADC_OVERFLOW': None,
        'H1:NO_DMT-ETMY_ESD_DAC_OVERFLOW': None}


def test_check_vectors_skips_mdc(caplog):
    """Test that detchar checks are skipped for MDC events."""
    caplog.set_level(logging.INFO)
    detchar.check_vectors({'search': 'MDC', 'graceid': 'M1234'}, 'S123', 0, 1)
    messages = [record.message for record in caplog.records]
    assert 'Skipping detchar checks because M1234 is an MDC' in messages


def test_check_vectors_skips_ew(caplog):
    """Test that detchar checks are skipped for events in the future."""
    caplog.set_level(logging.INFO)
    detchar.check_vectors({'pipeline': 'gstlal', 'graceid': 'S1234'},
                          'S123', 99999999999999999, 99999999999999999.5)
    messages = [record.message for record in caplog.records]
    assert 'Skipping detchar checks because S1234 is in the future' in messages  # noqa: E501


@patch('gwcelery.tasks.detchar.dqr_json', return_value='dqrjson')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.gracedb.remove_label')
@patch('gwcelery.tasks.gracedb.create_label')
def test_check_vectors(mock_create_label, mock_remove_label, mock_upload,
                       mock_json, llhoft_glob_pass, ifo_h1, ifo_h1_idq):
    event = {'search': 'AllSky', 'instruments': 'H1', 'pipeline': 'oLIB'}
    superevent_id = 'S12345a'
    start, end = 1216577978, 1216577978.1
    detchar.check_vectors(event, superevent_id, start, end)
    calls = [
        call(
            None, None, 'S12345a',
            ('Detector state for active instruments is good.\n{}'
             'Check looked within -1.5/+1.5 seconds of superevent. ').format(
                 detchar.generate_table(
                     'Data quality bits',
                     ['H1:NO_OMC_DCPD_ADC_OVERFLOW',
                      'H1:NO_DMT-ETMY_ESD_DAC_OVERFLOW',
                      'H1:HOFT_OK', 'H1:OBSERVATION_INTENT'], [], [])),
            ['data_quality']),
        call(
            None, None, 'S12345a',
            ('No HW injections found. '
             'Check looked within -1.5/+1.5 seconds of superevent. '),
            ['data_quality']),
        call(
            None, None, 'S12345a',
            ('iDQ false alarm probabilities at both H1 and L1'
             ' are good (above {} threshold). '
             'Minimum FAP is "H1:IDQ-FAP_OVL_32_2048": 1.0. '
             'Check looked within -1.5/+1.5 seconds of superevent. ').format(
                 app.conf['idq_fap_thresh']),
            ['data_quality']),
        call(
            '"dqrjson"', 'gwcelerydetcharcheckvectors-S12345a.json', 'S12345a',
            'DQR-compatible json generated from check_vectors results'),
    ]
    mock_upload.assert_has_calls(calls, any_order=True)
    mock_create_label.assert_called_with('DQOK', 'S12345a')
    mock_remove_label.assert_has_calls(
        [call('DQV', 'S12345a'),
         call('INJ', 'S12345a')],
        any_order=True)


@patch('gwcelery.tasks.detchar.dqr_json', return_value='dqrjson')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.gracedb.remove_label')
@patch('gwcelery.tasks.gracedb.create_label')
def test_gatedhoft_skips_dmtvec(mock_create_label, mock_remove_label,
                                mock_upload, mock_json, llhoft_glob_pass,
                                ifo_h1, ifo_h1_idq, gatedpipe,
                                gatedpipe_prepost):
    event = {'search': 'AllSky', 'instruments': 'H1', 'pipeline': 'gatepipe'}
    superevent_id = 'S12345a'
    start, end = 1216577977, 1216577979
    detchar.check_vectors(event, superevent_id, start, end)
    mock_upload.assert_has_calls([
            call(None, None, 'S12345a',
                 ('Detector state for active instruments is good.\n{}'
                  'Check looked within -0.5/+0.5 seconds of superevent. '
                  'Pipeline gatepipe uses gated h(t), '
                  'LIGO DMT-DQ_VECTOR not checked.').format(
                      detchar.generate_table(
                          'Data quality bits',
                          ['H1:HOFT_OK', 'H1:OBSERVATION_INTENT'], [], [])),
                 ['data_quality']), ], any_order=True)
    mock_remove_label.assert_has_calls(
        [call('DQV', superevent_id),
         call('INJ', superevent_id)],
        any_order=True)
    mock_create_label.assert_called_once_with('DQOK', superevent_id)


def test_detchar_generate_table():
    result = ('<table '
              'style="width:100%;border-collapse:collapse;text-align:center;"'
              ' border="1">\n<th'
              ' colspan="2">title</th>\n<tr><th>Bit</th><th>Value</th></tr>\n'
              '    <tr><td>good_bit_1</td><td bgcolor="#0f0">1</td></tr>\n'
              '    <tr><td>good_bit_2</td><td bgcolor="#0f0">1</td></tr>\n'
              '    <tr><td>bad_bit_1</td><td bgcolor="red">0</td></tr>\n'
              '    <tr><td>bad_bit_2</td><td bgcolor="red">0</td></tr>\n'
              '    <tr><td>unknown_bit_1</td>'
              '<td bgcolor="yellow">unknown</td></tr>\n'
              '    <tr><td>unknown_bit_2</td>'
              '<td bgcolor="yellow">unknown</td></tr>\n'
              '    <tr bgcolor="red"><th>Overall state for all '
              'detectors</th><th>bad</th></tr>\n'
              '</table>')
    assert detchar.generate_table(
        'title', ['good_bit_1', 'good_bit_2'],
        ['bad_bit_1', 'bad_bit_2'], ['unknown_bit_1', 'unknown_bit_2']
    ) == result



"""Flask web application for manually triggering certain tasks."""
from contextlib import ExitStack, nullcontext

from celery.bin.base import CeleryDaemonCommand, CeleryOption, LOG_LEVEL
from celery.platforms import detached
import click
from flask.cli import FlaskGroup

from ..flask import app
from .. import views as _  # noqa: F401


class CeleryDaemonFlaskGroup(FlaskGroup):

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.params.extend(CeleryDaemonCommand(name='flask').params)


def maybe_detached(*, detach, **kwargs):
    return detached(**kwargs) if detach else nullcontext()


@click.group(cls=CeleryDaemonFlaskGroup, help=__doc__,
             context_settings={'auto_envvar_prefix': 'FLASK'})
@click.option('-D',
              '--detach',
              cls=CeleryOption,
              is_flag=True,
              default=False,
              help="Start as a background process.")
@click.option('-l',
              '--loglevel',
              default='WARNING',
              cls=CeleryOption,
              type=LOG_LEVEL,
              help="Logging level.")
@click.pass_context
def flask(ctx, detach=None, loglevel=None, logfile=None,
          pidfile=None, uid=None, gid=None, umask=None, **kwargs):
    # Look up Celery app
    celery_app = ctx.parent.obj.app

    # # Prepare to pass Flask app to Flask CLI
    ctx.obj.create_app = lambda *args, **kwargs: app

    # Detach from the tty if requested.
    # FIXME: After we update to click >= 8, replace the elaborate construction
    # below with ctx.with_resource(maybe_detached(...))
    exit_stack = ExitStack()
    ctx.call_on_close(exit_stack.close)
    exit_stack.enter_context(maybe_detached(detach=detach,
                                            logfile=logfile, pidfile=pidfile,
                                            uid=uid, gid=gid, umask=umask))

    celery_app.log.setup(loglevel, logfile)



from unittest.mock import call, patch

import pytest

from .test_tasks_skymaps import toy_fits_filecontents  # noqa: F401
from ..tasks import gracedb
from ..tasks import raven


@pytest.mark.live_worker
@pytest.mark.parametrize(
    'group,gracedb_id,pipelines,ext_search,se_search,tl,th',
    [['CBC', 'S1', ['Fermi', 'Swift'], [], ['GRB'], -1, 5],
     ['Burst', 'S2', ['Fermi', 'Swift'], [], ['GRB'], -60, 600],
     ['Burst', 'S3', ['SNEWS'], ['Supernova'], [], -10, 10],
     ['Burst', 'T4', ['SNEWS'], 'Supernova', [], -10, 10],
     ['CBC', 'MS4', ['Fermi'], ['MDC'], [], -1, 5],
     ['CBC', 'E1', ['Fermi'], [], ['GRB'], -5, 1],
     ['CBC', 'M1', ['Fermi'], [], ['MDC'], -5, 1]])
@patch('gwcelery.tasks.gracedb.create_label')
@patch('gwcelery.tasks.raven.raven_pipeline.s')
@patch('gwcelery.tasks.raven.search.si', return_value=[{'superevent_id': 'S5',
                                                        'graceid': 'E2'}])
@patch('gwcelery.tasks.raven.calculate_coincidence_far')
def test_coincidence_search(mock_calculate_coincidence_far,
                            mock_search, mock_raven_pipeline,
                            mock_create_label,
                            group, gracedb_id, pipelines,
                            ext_search, se_search, tl, th):
    """Test that correct time windows are used for each RAVEN search."""
    alert_object = {'superevent_id': gracedb_id}
    if 'E' in gracedb_id:
        alert_object['group'] = 'External'
    raven.coincidence_search(gracedb_id, alert_object, group,
                             pipelines, ext_search, se_search)

    mock_search.assert_called_once_with(
        gracedb_id, alert_object, tl, th, group, pipelines,
        ext_search, se_search)
    mock_raven_pipeline.assert_called_once_with(
        gracedb_id, alert_object, tl, th, group)


@pytest.mark.parametrize(
    'event_type,event_id', [['SE', 'S1234'], ['ExtTrig', 'E1234']])
@patch('ligo.raven.search.search')
def test_raven_search(mock_raven_search, event_type, event_id):
    """Test that correct input parameters are used for raven."""
    alert_object = {}
    if event_type == 'SE':
        alert_object['superevent_id'] = event_id

    # call raven search
    raven.search(event_id, alert_object)
    mock_raven_search.assert_called_once_with(
        event_id, -5, 5, event_dict=alert_object,
        gracedb=gracedb.client, group=None, pipelines=[], searches=[],
        se_searches=[])


@pytest.mark.parametrize('group', ['CBC', 'Burst'])
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_coincidence_far(
        mock_calc_signif, group):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': [],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        se_dict=se, ext_dict=ext,
        incl_sky=False, grb_search='GRB',
        gracedb=gracedb.client, far_grb=None)


@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_coincidence_far_subgrb(mock_calc_signif):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': [],
           'far': 1e5}
    tl, th = -1, 10
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        se_dict=se, ext_dict=ext,
        incl_sky=False, grb_search='GRB',
        gracedb=gracedb.client, far_grb=1e5)


@pytest.mark.parametrize('group', ['CBC', 'Burst'])  # noqa: F811
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='fermi_skymap.fits.gz')
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_spacetime_coincidence_far_fermi(
        mock_calc_signif, mock_get_skymap_filename, group):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': ['EXT_SKYMAP_READY', 'SKYMAP_READY'],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=True, grb_search='GRB',
        se_dict=se, ext_dict=ext,
        se_fitsfile='fermi_skymap.fits.gz',
        ext_fitsfile='fermi_skymap.fits.gz',
        se_moc=True, ext_moc=False,
        use_radec=False,
        gracedb=gracedb.client, far_grb=None,
        use_preferred_event_skymap=False)


@pytest.mark.parametrize('group', ['CBC', 'Burst'])  # noqa: F811
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='swift_skymap.fits.gz')
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_spacetime_coincidence_far_swift(
        mock_calc_signif, mock_get_skymap_filename, group):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Swift',
           'search': 'GRB',
           'labels': ['EXT_SKYMAP_READY', 'SKYMAP_READY'],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=True, grb_search='GRB',
        se_dict=se, ext_dict=ext,
        se_fitsfile='swift_skymap.fits.gz',
        ext_fitsfile='swift_skymap.fits.gz',
        se_moc=True, ext_moc=False,
        use_radec=True,
        gracedb=gracedb.client, far_grb=None,
        use_preferred_event_skymap=False)


@pytest.mark.parametrize('group', ['CBC', 'Burst'])  # noqa: F811
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='fermi_skymap.fits.gz')
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_spacetime_coincidence_far_preferred(
        mock_calc_signif, mock_get_skymap_filename, group):
    se = {'superevent_id': 'S1234',
          'preferred_event': 'G1'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': ['EXT_SKYMAP_READY', 'EM_READY'],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=True, grb_search='GRB',
        se_dict=se, ext_dict=ext,
        se_fitsfile='fermi_skymap.fits.gz',
        ext_fitsfile='fermi_skymap.fits.gz',
        se_moc=True, ext_moc=False,
        use_radec=False,
        gracedb=gracedb.client, far_grb=None,
        use_preferred_event_skymap=True)


def mock_get_labels(superevent_id):
    if superevent_id == 'S14':
        return {'ADVREQ'}
    else:
        return {}


def mock_coinc_far(*args):
    return {'temporal_coinc_far': 1e-7,
            'spatiotemporal_coinc_far': None,
            'overlap_integral': None}


def mock_get_superevent(superevent_id):
    if superevent_id == 'S1':
        old_time_far = None
        old_space_far = None
    elif superevent_id == 'S2':
        old_time_far = 1e-5
        old_space_far = None
    elif superevent_id == 'S3':
        old_time_far = 1e-4
        old_space_far = None
    elif superevent_id == 'S4':
        old_time_far = 1e-5
        old_space_far = 1e-6
    elif superevent_id == 'S5':
        old_time_far = 1e-4
        old_space_far = 1e-6
    else:
        old_time_far = 1e-5
        old_space_far = None
    return {'time_coinc_far': old_time_far,
            'space_coinc_far': old_space_far,
            'superevent_id': superevent_id,
            'labels': [],
            'em_type': None}


@pytest.mark.parametrize(
    'raven_search_results,graceid,tl,th,group',
    [[[{'graceid': 'E1', 'pipeline': 'GRB'}], 'S1', -5, 1, 'CBC'],
     [[{'superevent_id': 'S10', 'far': 1, 'preferred_event': 'G1'}],
        'E2', -1, 5, 'CBC'],
     [[{'graceid': 'E3', 'pipeline': 'GRB'},
       {'graceid': 'E4', 'pipeline': 'GRB'}], 'S2', -600, 60, 'Burst'],
     [[{'superevent_id': 'S11', 'far': 1, 'preferred_event': 'G2'},
       {'superevent_id': 'S12', 'far': .001, 'preferred_event': 'G3'}],
        'E5', -1, 5, 'CBC'],
     [[], 'S13', -1, 5, 'CBC'],
     [[{'graceid': 'E4', 'pipeline': 'GRB'}], 'S14', -1, 5, 'CBC'],
     [[{'superevent_id': 'S12', 'far': .001, 'preferred_event': 'G3'}],
        'T4', -10, 10, 'Burst'],
     [[{'superevent_id': 'MS13', 'far': 1, 'preferred_event': 'M3'}],
        'M15', -1, 5, 'CBC'],
     [[{'superevent_id': 'S13', 'far': 1, 'preferred_event': 'G4'}],
        'E6', -60, 600, 'Burst']])
@patch('gwcelery.tasks.raven.trigger_raven_alert.run')
@patch('gwcelery.tasks.gracedb.get_labels', mock_get_labels)
@patch('gwcelery.tasks.raven.calculate_coincidence_far.run',
       return_value=mock_coinc_far())
@patch('gwcelery.tasks.gracedb.update_superevent')
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.external_skymaps.plot_overlap_integral.run')
@patch('gwcelery.tasks.gracedb.get_superevent', mock_get_superevent)
def test_raven_pipeline(mock_plot_overlap_integral,
                        mock_create_label,
                        mock_update_superevent,
                        mock_calculate_coincidence_far,
                        mock_trigger_raven_alert,
                        raven_search_results, graceid, tl, th, group):
    """This function tests that the RAVEN pipeline runs correctly for scenarios
    where RAVEN finds nothing, a coincidence is found but does not pass
    threshold, when a coincidence is found but does pass threshold, and when
    multiple events are found.
    """
    alert_object = {'preferred_event': 'G1', 'pipeline': 'Fermi',
                    'search': 'Supernova' if graceid == 'T4' else 'GRB',
                    'labels': [],
                    'superevent': None if graceid != 'E6' else 'S14'}
    time_coinc_far = mock_coinc_far()['temporal_coinc_far']
    space_coinc_far = mock_coinc_far()['spatiotemporal_coinc_far']

    for result in raven_search_results:
        result['labels'] = []
    if 'S' not in graceid:
        alert_object['group'] = 'External'
        if 'T' in graceid:
            alert_object['group'] = 'Test'
        alert_object['graceid'] = graceid
        for result in raven_search_results:
            result['time_coinc_far'] = 1e-5
            result['space_coinc_far'] = None
            result['overlap_integral'] = None
    else:
        alert_object['superevent_id'] = graceid
        alert_object['time_coinc_far'] = 1e-5
        alert_object['space_coinc_far'] = None
        alert_object['overlap_integral'] = None
        for result in raven_search_results:
            result['superevent'] = None
    raven.raven_pipeline(raven_search_results, graceid, alert_object, tl, th,
                         group)

    coinc_calls = []
    label_calls = []
    update_calls = []
    plot_calls = []
    if not raven_search_results:
        mock_calculate_coincidence_far.assert_not_called()
        mock_create_label.assert_not_called()
        return
    if 'S' in graceid:
        for result in raven_search_results:
            label_calls.append(call('EM_COINC', result['graceid']))
            coinc_calls.append(call(alert_object, result, tl, th))
            label_calls.append(call('EM_COINC', graceid))
            update_calls.append(
                call(graceid, em_type=result['graceid'],
                     time_coinc_far=time_coinc_far,
                     space_coinc_far=space_coinc_far))
            plot_calls.append(
                call(mock_coinc_far(), alert_object, result))
    else:
        result = raven.preferred_superevent(raven_search_results)[0]
        label_calls.append(call('EM_COINC', result['superevent_id']))
        coinc_calls.append(call(result, alert_object, tl, th))
        label_calls.append(call('EM_COINC', graceid))
        update_calls.append(
            call(raven_search_results[-1]['superevent_id'],
                 em_type=graceid,
                 time_coinc_far=time_coinc_far,
                 space_coinc_far=space_coinc_far))
        plot_calls.append(
            call(mock_coinc_far(), result, alert_object))

    alert_calls = []
    if 'S' in graceid:
        for result in raven_search_results:
            alert_calls.append(call(mock_coinc_far(),
                                    alert_object, graceid, result, group))
    else:
        alert_calls.append(call(
            mock_coinc_far(),
            result, graceid, alert_object, group))
    if graceid != 'E6':
        mock_trigger_raven_alert.assert_has_calls(alert_calls, any_order=True)

        mock_calculate_coincidence_far.assert_has_calls(coinc_calls,
                                                        any_order=True)
        mock_update_superevent.assert_has_calls(update_calls, any_order=True)
        mock_create_label.assert_has_calls(label_calls, any_order=True)
        mock_plot_overlap_integral.assert_has_calls(plot_calls, any_order=True)
    else:
        mock_trigger_raven_alert.assert_not_called()
        mock_calculate_coincidence_far.assert_not_called()
        mock_update_superevent.assert_not_called()
        mock_create_label.assert_not_called()
        mock_plot_overlap_integral.assert_not_called()


@pytest.mark.parametrize(
    'raven_search_results, testnum',
    [[[{'superevent_id': 'S10', 'far': 1, 'preferred_event': 'G1'}], 1],
     [[{'superevent_id': 'S11', 'far': 1, 'preferred_event': 'G2'},
       {'superevent_id': 'S12', 'far': .001, 'preferred_event': 'G3'}], 2],
     [[{'superevent_id': 'S13', 'far': 1, 'preferred_event': 'G4'},
       {'superevent_id': 'S14', 'far': .0001, 'preferred_event': 'G5'},
       {'superevent_id': 'S15', 'far': .001, 'preferred_event': 'G6'}], 3]])
def test_preferred_superevent(raven_search_results, testnum):

    preferred_superevent = raven.preferred_superevent(raven_search_results)
    if testnum == 1:
        assert preferred_superevent == [{'superevent_id': 'S10', 'far': 1,
                                         'preferred_event': 'G1'}]
    if testnum == 2:
        assert preferred_superevent == [{'superevent_id': 'S12', 'far': .001,
                                         'preferred_event': 'G3'}]
    if testnum == 3:
        assert preferred_superevent == [{'superevent_id': 'S14', 'far': .0001,
                                         'preferred_event': 'G5'}]


@pytest.mark.parametrize(
    'new_time_far,new_space_far,superevent_id,pipeline,result',
    [[1e-4, None, 'S1', 'Fermi', True],
     [1e-4, 1e-3, 'S1', 'Swift', True],
     [1e-4, None, 'S2', 'AGILE', False],
     [1e-4, 1e-3, 'S3', 'Fermi', True],
     [1e-4, 1e-3, 'S4', 'Fermi', False],
     [1e-8, None, 'S4', 'Swift', False],
     [1e-4, 1e-8, 'S5', 'AGILE', True],
     [None, None, 'S1', 'SNEWS', True]])
@patch('gwcelery.tasks.gracedb.update_superevent')
@patch('gwcelery.tasks.gracedb.get_superevent', mock_get_superevent)
def test_update_superevent(mock_update_superevent,
                           new_time_far, new_space_far,
                           superevent_id,
                           pipeline, result):
    """This function tests that new coincidences update the superevent when
    they have a superior joint FAR or are a SNEWS event.
    """
    superevent = {'superevent_id': superevent_id,
                  'labels': []}
    ext_event = {'graceid': 'E100',
                 'pipeline': pipeline}
    coinc_far_dict = {'temporal_coinc_far': new_time_far,
                      'spatiotemporal_coinc_far': new_space_far}
    raven.update_coinc_far(coinc_far_dict, superevent, ext_event)
    if result:
        mock_update_superevent.assert_called_with(
            superevent_id, em_type='E100',
            time_coinc_far=new_time_far,
            space_coinc_far=new_space_far)
    else:
        mock_update_superevent.assert_not_called()


def mock_get_event(graceid):
    if 'E' in graceid:
        if graceid == "E1":
            pipeline, search, labels = 'Fermi', 'GRB', []
        elif graceid == "E2":
            pipeline, search, labels = 'Fermi', 'GRB', ['NOT_GRB']
        elif graceid == "E3":
            pipeline, search, labels = 'SNEWS', 'Supernova', []
        elif graceid == "E4":
            pipeline, search, labels = 'Fermi', 'GRB', []
        elif graceid == "E5":
            pipeline, search, labels = 'Fermi', 'GRB', ['RAVEN_ALERT']
        return {'superevent_id': graceid.replace('E', 'S'),
                'graceid': graceid,
                'pipeline': pipeline,
                'search': search,
                'labels': labels}
    elif 'S' in graceid:
        return {'em_type': graceid.replace('S', 'E'),
                'time_coinc_far': 1e-20,
                'space_coinc_far': 1e-20,
                'superevent_id': graceid,
                'labels': ['COMBINEDSKYMAP_READY'] if graceid == 'S4' else []}


@pytest.mark.parametrize(
    'superevent_id,ext_event_id_old,ext_event_id_new,result',
    # NOT_GRB should not supersede real GRB
    [['S1', 'E1', 'E2', False],
     # Real GRB should supersede NOT_GRB
     ['S2', 'E2', 'E1', True],
     # SNEWS event should not be overwritten by GRB
     ['S3', 'E3', 'E1', False],
     # RAVEN_ALERT should supersede subthreshold joint candidate
     ['S4', 'E4', 'E5', True],
     # Subthreshold joint candidate should not overwrite RAVEN_ALERT
     ['S5', 'E5', 'E4', False]]
)
@patch('gwcelery.tasks.gracedb.update_superevent')
@patch('gwcelery.tasks.gracedb.get_event', mock_get_event)
@patch('gwcelery.tasks.gracedb.get_superevent', mock_get_event)
def test_update_superevent_w_emtype(mock_update_superevent,
                                    superevent_id,
                                    ext_event_id_old, ext_event_id_new,
                                    result):
    """This function tests that new coincidences update the superevent
    if the new GRB is likely a real GRB and the old is likely not.
    Also tested is that SNEWS event aren't overwritten by GRBs and locked
    events are overwritten, denote by the COMBINEDSKYMAP_READY label in the
    superevent which is applied just before sending an alert.
    """
    superevent = mock_get_event(superevent_id)
    superevent['em_type'] = ext_event_id_old
    ext_event_new = mock_get_event(ext_event_id_new)
    ext_event_old = mock_get_event(ext_event_id_old)
    time_coinc_far = 1e-10 if ext_event_old['labels'] else 1e-20
    space_coinc_far = 1e-10 if ext_event_old['labels'] else 1e-30
    coinc_far_dict = {'temporal_coinc_far': time_coinc_far,
                      'spatiotemporal_coinc_far': space_coinc_far}
    raven.update_coinc_far(coinc_far_dict, superevent, ext_event_new)
    if result:
        mock_update_superevent.assert_called_with(
            superevent_id,
            em_type=ext_event_id_new if result else ext_event_id_old,
            time_coinc_far=time_coinc_far,
            space_coinc_far=space_coinc_far)
    else:
        mock_update_superevent.assert_not_called()


def _mock_get_event(graceid):
    if graceid == "S1234":
        return {"superevent_id": "S1234",
                "preferred_event": "G000001",
                "preferred_event_data": {"group": "Burst"},
                "far": 1e-5}
    elif graceid == "S2345":
        return {"superevent_id": "S2345",
                "preferred_event": "G000002",
                "preferred_event_data": {"group": "Burst"},
                "far": 1e-10}
    elif graceid == "S2468":
        return {"superevent_id": "S2468",
                "preferred_event": "G000002",
                "preferred_event_data": {"group": "CBC"},
                "far": 1e-4}
    elif graceid == "S5678":
        return {"superevent_id": "S5678",
                "preferred_event": "G000003",
                "preferred_event_data": {"group": "CBC"},
                "far": 1e-5}
    elif graceid == "S8642":
        return {"superevent_id": "S8642",
                "preferred_event": "G000003",
                "preferred_event_data": {"group": "Burst"},
                "far": 1e-3}
    elif graceid == "S9876":
        return {"superevent_id": "S9876",
                "preferred_event": "G000003",
                "preferred_event_data": {"group": "Burst"},
                "far": 1e-6}
    elif graceid == "TS1111":
        return {"superevent_id": "TS1111",
                "preferred_event": "G000003",
                "preferred_event_data": {"group": "Test"},
                "far": 1e-6}
    elif graceid == "MS1111":
        return {"superevent_id": "MS1111",
                "preferred_event": "M000004",
                "preferred_event_data": {"group": "CBC"},
                "far": 1e-9}
    elif graceid == 'E1':
        return {"graceid": "E1",
                "pipeline": 'Swift',
                "search": 'GRB',
                "labels": [],
                "group": 'External'}
    elif graceid == 'E2':
        return {"graceid": "E2",
                "pipeline": 'Fermi',
                "search": 'SubGRB',
                "labels": [],
                "group": 'External'}
    elif graceid == 'E3':
        return {"graceid": "E3",
                "pipeline": 'SNEWS',
                "search": 'Supernova',
                "labels": [],
                "group": 'External'}
    elif graceid == 'T4':
        return {"graceid": "T4",
                "pipeline": 'SNEWS',
                "search": 'Supernova',
                "labels": [],
                "group": 'Test'}
    elif graceid == 'E5':
        return {"graceid": "E5",
                "pipeline": 'Fermi',
                "search": 'GRB',
                "labels": ['NOT_GRB'],
                "group": 'External'}
    elif graceid == 'M6':
        return {"graceid": "M6",
                "pipeline": 'Fermi',
                "search": 'MDC',
                "labels": [],
                "group": 'External'}
    elif graceid == 'E7':
        return {"graceid": "E7",
                "pipeline": 'Fermi',
                "search": 'GRB',
                "labels": [],
                "group": 'External'}
    else:
        raise AssertionError


def _mock_get_labels(graceid):
    return _mock_get_event(graceid)['labels']


def _mock_get_coinc_far(graceid):
    if graceid == "S1234":
        return {"temporal_coinc_far": 1e-12,
                "spatiotemporal_coinc_far": 1e-05}
    elif graceid == "S2468":
        return {"temporal_coinc_far": 1e-09,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S5678":
        return {"temporal_coinc_far": 1e-15,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S8642":
        return {"temporal_coinc_far": 1e-03,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S9876":
        return {"temporal_coinc_far": 1e-07,
                "spatiotemporal_coinc_far": 1e-13}
    elif graceid == "MS1111":
        return {"temporal_coinc_far": 1e-09,
                "spatiotemporal_coinc_far": 1e-10}
    else:
        return {}


@pytest.mark.parametrize(
    'graceid,result_id,group,expected_result',
    [['S1234', 'E1', 'Burst', False],
     ['S2468', 'E1', 'CBC', False],
     ['S2468', 'E5', 'CBC', False],
     ['S2468', 'E2', 'CBC', False],
     ['S2345', 'E3', 'Burst', True],
     ['MS1111', 'M6', 'CBC', True],
     ['S5678', 'E1', 'CBC', False],
     ['S5678', 'E7', 'CBC', False],
     ['E1', 'S9876', 'CBC', True],
     ['E1', 'S2468', 'CBC', False],
     ['E2', 'S5678', 'CBC', False],
     ['E3', 'S8642', 'Burst', False],
     ['E3', 'S9876', 'Burst', True],
     ['T4', 'TS1111', 'CBC', False],
     ['E5', 'S5678', 'CBC', False]])
@patch('gwcelery.tasks.gracedb.get_labels', side_effect=_mock_get_labels)
@patch('gwcelery.tasks.gracedb.update_superevent')
@patch('gwcelery.tasks.gracedb.create_label.run')
def test_trigger_raven_alert(mock_create_label, mock_update_superevent,
                             mock_get_labels,
                             graceid, result_id, group, expected_result):
    if 'S' in graceid:
        superevent_id = graceid
        ext_id = result_id
    else:
        superevent_id = result_id
        ext_id = graceid
    superevent = _mock_get_event(superevent_id)
    coinc_far_json = _mock_get_coinc_far(superevent_id)
    ext_event = _mock_get_event(ext_id)
    preferred_id = superevent['preferred_event']
    raven.trigger_raven_alert(coinc_far_json, superevent,
                              graceid, ext_event, group)

    if expected_result:
        label_calls = [call('RAVEN_ALERT', superevent_id),
                       call('HIGH_PROFILE', superevent_id),
                       call('RAVEN_ALERT', ext_id),
                       call('RAVEN_ALERT', preferred_id)]
        mock_create_label.assert_has_calls(label_calls)
    else:
        mock_create_label.assert_not_called()


@pytest.mark.parametrize(
    'se_id,ext_id,expected_result',
    [['S2468', 'E5', True],
     ['S2468', 'E1', True],
     ['S2468', 'E2', False],
     ['S2468', 'M6', True],
     ['S2345', 'E1', False],
     ['S2345', 'E2', False]])
@patch('gwcelery.tasks.gracedb.create_label.run')
def test_sog_paper_pipeline(mock_create_label,
                            se_id, ext_id, expected_result):

    superevent = _mock_get_event(se_id)
    ext_event = _mock_get_event(ext_id)
    if se_id == 'S2468':
        superevent['far'] = 1e-12
        superevent['space_coinc_far'] = 1e-13
    elif se_id == 'S2345':
        superevent['far'] = 1e-14
        superevent['space_coinc_far'] = 1e-8
    raven.sog_paper_pipeline(ext_event, superevent)

    if expected_result:
        mock_create_label.assert_called_once_with(
            'SOG_READY', superevent['superevent_id'])
    else:
        mock_create_label.assert_not_called()



import datetime
from unittest.mock import MagicMock, Mock, patch, call

from flask import get_flashed_messages, url_for
from requests.exceptions import HTTPError
from requests.models import Response
import pytest
from werkzeug.http import HTTP_STATUS_CODES

from .. import flask
from .. import views as _  # noqa: F401


@pytest.fixture
def app():
    return flask.app


def test_send_preliminary_gcn_post_no_data(client, monkeypatch):
    """Test send_update_gcn endpoint with no form data."""
    mock_update_superevent = Mock()
    mock_get_event = Mock()
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.update_superevent.run',
        mock_update_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.orchestrator.earlywarning_preliminary_alert.run',
        mock_preliminary_alert)

    response = client.post(url_for('send_preliminary_gcn'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'No alert sent. Please fill in all fields.']
    mock_update_superevent.assert_not_called()
    mock_get_event.assert_not_called()
    mock_preliminary_alert.assert_not_called()


def test_send_preliminary_gcn_post(client, monkeypatch):
    """Test send_update_gcn endpoint with complete form data."""
    mock_update_superevent = Mock()
    mock_event = Mock()
    mock_superevent = Mock()
    mock_get_event = Mock(return_value=mock_event)
    mock_get_superevent = Mock(return_value=mock_superevent)
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.update_superevent.run',
        mock_update_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_superevent.run',
        mock_get_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.orchestrator.earlywarning_preliminary_alert.run',
        mock_preliminary_alert)

    response = client.post(url_for('send_preliminary_gcn'), data={
        'superevent_id': 'MS190208a',
        'event_id': 'M12345'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'Queued preliminary alert for MS190208a.']
    mock_update_superevent.assert_called_once_with(
        'MS190208a', preferred_event='M12345')
    mock_get_event.assert_called_once_with('M12345')
    mock_get_superevent.assert_called_once_with('MS190208a')

    alert = {
        'uid': 'MS190208a',
        'object': mock_superevent
    }
    mock_preliminary_alert.assert_called_once_with(mock_event, alert,
                                                   alert_type='preliminary',
                                                   initiate_voevent=True)


def test_change_preferred_event_post(client, monkeypatch):
    """Test changing preferred event with complete form data."""
    mock_update_superevent = Mock()
    mock_add_pipeline_pref_event = Mock()
    mock_event = MagicMock()
    mock_superevent = MagicMock()
    event_dict_spoofer = {
        'pipeline': 'gstlal',
    }
    superevent_dict_spoofer = {
        'pipeline_preferred_events': {'gstlal': {'graceid': 'M12345'}}
    }
    mock_event.__getitem__.side_effect = event_dict_spoofer.__getitem__
    mock_superevent.__getitem__.side_effect = \
        superevent_dict_spoofer.__getitem__
    mock_get_event = Mock(return_value=mock_event)
    mock_get_superevent = Mock(return_value=mock_superevent)
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.update_superevent.run',
        mock_update_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_superevent.run',
        mock_get_superevent)
    monkeypatch.setattr(
        'gwcelery.views._construct_igwn_alert_and_send_prelim_alert.run',
        mock_preliminary_alert)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.add_pipeline_preferred_event.run',
        mock_add_pipeline_pref_event)

    response = client.post(url_for('change_preferred_event'), data={
        'superevent_id': 'MS190208a',
        'event_id': 'M12345'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'Changed preferred event for MS190208a.']
    mock_update_superevent.assert_called_once_with(
        'MS190208a', preferred_event='M12345')
    mock_get_event.assert_called_once_with('M12345')
    mock_get_superevent.assert_called_once_with('MS190208a')
    mock_preliminary_alert.assert_called_once_with([mock_superevent,
                                                   mock_event], 'MS190208a',
                                                   initiate_voevent=False)
    # This should only be called if the new preferred event is not already a
    # pipeline preferred event
    mock_add_pipeline_pref_event.assert_not_called()


def test_change_preferred_event_pipeline_preferred_event_post(client,
                                                              monkeypatch):
    """Test changing preferred event to event that isn't a pipeline preferred
    event with complete form data."""
    mock_update_superevent = Mock()
    mock_add_pipeline_pref_event = Mock()
    mock_event = MagicMock()
    mock_superevent = MagicMock()
    event_dict_spoofer = {'pipeline': 'gstlal'}
    superevent_dict_spoofer = {
        'pipeline_preferred_events': {'gstlal': {'graceid': 'M12344'}}
    }
    mock_event.__getitem__.side_effect = event_dict_spoofer.__getitem__
    mock_superevent.__getitem__.side_effect = \
        superevent_dict_spoofer.__getitem__
    mock_get_event = Mock(return_value=mock_event)
    mock_get_superevent = Mock(return_value=mock_superevent)
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.update_superevent.run',
        mock_update_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_superevent.run',
        mock_get_superevent)
    monkeypatch.setattr(
        'gwcelery.views._construct_igwn_alert_and_send_prelim_alert.run',
        mock_preliminary_alert)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.add_pipeline_preferred_event.run',
        mock_add_pipeline_pref_event)

    response = client.post(url_for('change_preferred_event'), data={
        'superevent_id': 'MS190208a',
        'event_id': 'M12345'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'Changed preferred event for MS190208a.']
    mock_update_superevent.assert_called_once_with(
        'MS190208a', preferred_event='M12345')
    mock_get_event.assert_called_once_with('M12345')
    mock_get_superevent.assert_called_once_with('MS190208a')
    mock_preliminary_alert.assert_called_once_with([mock_superevent,
                                                   mock_event], 'MS190208a',
                                                   initiate_voevent=False)
    # This should only be called if the new preferred event is not already a
    # pipeline preferred event
    mock_add_pipeline_pref_event.assert_called_once_with('MS190208a', 'M12345')


def test_successful_change_pipeline_preferred_event_post(client, monkeypatch):
    """Test changing pipeline preferred event with complete form data."""
    mock_add_pipeline_pref_event = Mock()
    mock_event = MagicMock()
    mock_superevent = MagicMock()
    event_dict_spoofer = {'pipeline': 'gstlal'}
    superevent_dict_spoofer = {
        'preferred_event_data': {'pipeline': 'mbta'},
        'pipeline_preferred_events': {'gstlal': {'graceid': 'M12344'}}
    }
    mock_event.__getitem__.side_effect = event_dict_spoofer.__getitem__
    mock_superevent.__getitem__.side_effect = \
        superevent_dict_spoofer.__getitem__
    mock_get_event = Mock(return_value=mock_event)
    mock_get_superevent = Mock(return_value=mock_superevent)
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_superevent.run',
        mock_get_superevent)
    monkeypatch.setattr(
        'gwcelery.views._construct_igwn_alert_and_send_prelim_alert.run',
        mock_preliminary_alert)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.add_pipeline_preferred_event.run',
        mock_add_pipeline_pref_event)

    response = client.post(url_for('change_pipeline_preferred_event'), data={
        'superevent_id': 'MS190208a',
        'pipeline': 'gstlal',
        'event_id': 'M12345'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'Changed gstlal preferred event for MS190208a.']
    mock_get_event.assert_called_once_with('M12345')
    mock_get_superevent.assert_called_once_with('MS190208a')
    mock_preliminary_alert.assert_called_once_with([mock_superevent,
                                                   mock_event], 'MS190208a',
                                                   initiate_voevent=False)
    mock_add_pipeline_pref_event.assert_called_once_with('MS190208a', 'M12345')


def test_rejected_change_pipeline_preferred_event_post(client, monkeypatch):
    """Test attempting to change a pipeline preferred event that is also the
    superevent's preferred event with complete form data."""
    mock_add_pipeline_pref_event = Mock()
    mock_event = MagicMock()
    mock_superevent = MagicMock()
    event_dict_spoofer = {'pipeline': 'gstlal'}
    superevent_dict_spoofer = {
        'preferred_event_data': {'pipeline': 'gstlal'},
        'pipeline_preferred_events': {'gstlal': {'graceid': 'M12344'}}
    }
    mock_event.__getitem__.side_effect = event_dict_spoofer.__getitem__
    mock_superevent.__getitem__.side_effect = \
        superevent_dict_spoofer.__getitem__
    mock_get_event = Mock(return_value=mock_event)
    mock_get_superevent = Mock(return_value=mock_superevent)
    mock_preliminary_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_superevent.run',
        mock_get_superevent)
    monkeypatch.setattr(
        'gwcelery.views._construct_igwn_alert_and_send_prelim_alert.run',
        mock_preliminary_alert)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.add_pipeline_preferred_event.run',
        mock_add_pipeline_pref_event)

    response = client.post(url_for('change_pipeline_preferred_event'), data={
        'superevent_id': 'MS190208a',
        'pipeline': 'gstlal',
        'event_id': 'M12345'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'No change performed. User specified pipeline, gstlal, is the same '
        'pipeline that produced MS190208a\'s preferred event.'
    ]
    mock_get_event.assert_called_once_with('M12345')
    mock_get_superevent.assert_called_once_with('MS190208a')
    mock_preliminary_alert.assert_not_called()
    mock_add_pipeline_pref_event.assert_not_called()


def test_send_update_gcn_get(client):
    """Test send_update_gcn endpoint with disallowed HTTP method."""
    # GET requests not allowed
    response = client.get(url_for('send_update_gcn'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Method Not Allowed'


def test_send_update_gcn_post_no_data(client):
    """Test send_update_gcn endpoint with no form data."""
    response = client.post(url_for('send_update_gcn'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'No alert sent. Please fill in all fields.']


def test_send_update_gcn_post(client, monkeypatch):
    """Test send_update_gcn endpoint with complete form data."""
    mock_update_alert = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.orchestrator.update_alert.run', mock_update_alert)

    response = client.post(url_for('send_update_gcn'), data={
        'superevent_id': 'MS190208a',
        'skymap_filename': 'bayestar.fits.gz',
        'em_bright_filename': 'em_bright.json',
        'p_astro_filename': 'p_astro.json'})

    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'Queued update alert for MS190208a.']
    mock_update_alert.assert_called_once_with(
        ['bayestar.fits.gz', 'em_bright.json', 'p_astro.json'], 'MS190208a')


def test_send_update_gcn_circular_post_no_data(client):
    """Test send_update_gcn_circular endpoint with no form data."""
    response = client.post(url_for('create_update_gcn_circular'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'No circular created. Please fill in superevent ID and at ' +
        'least one update type.']


@pytest.mark.parametrize(
     'sky_loc,em_bright,p_astro,raven,answer',
     [["True", None, None, None, ['sky_localization']],
      [None, "True", "True", None, ['em_bright', 'p_astro']],
      [None, None, None, "True", ['raven']],
      ["True", "True", "True", "True",
       ['sky_localization', 'em_bright', 'p_astro', 'raven']]])
@patch('gwcelery.tasks.circulars.create_update_circular', return_value='')
def test_send_update_gcn_circular_post(mock_create_circular,
                                       sky_loc, em_bright, p_astro, raven,
                                       answer,
                                       client):
    """Test send_update_gcn_circular endpoint with complete form data."""
    response = client.post(url_for('create_update_gcn_circular'), data={
        'superevent_id': 'MS190208a',
        'sky_localization': sky_loc,
        'em_bright': em_bright,
        'p_astro': p_astro,
        'raven': raven})

    assert HTTP_STATUS_CODES[response.status_code] == 'OK'
    mock_create_circular.assert_called_once_with(
        'MS190208a', update_types=answer)


def test_typeahead_superevent_id(client, monkeypatch):
    """Test typeahead filtering for superevent_id."""
    mock_superevents = Mock(return_value=(
        {
            'superevent_id': (
                datetime.date(2019, 2, 1) + datetime.timedelta(i)
            ).strftime('MS%y%m%da')
        } for i in range(31)))
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.client.superevents.search', mock_superevents)

    response = client.get(
        url_for('typeahead_superevent_id', superevent_id='MS1902'))

    assert HTTP_STATUS_CODES[response.status_code] == 'OK'
    assert response.json == [
        'MS190201a', 'MS190202a', 'MS190203a', 'MS190204a', 'MS190205a',
        'MS190206a', 'MS190207a', 'MS190208a']


def test_typeahead_superevent_id_invalid_date(client, monkeypatch):
    """Test typeahead filtering for superevent_id when the search term contains
    an invalid date fragment.
    """
    mock_superevents = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.client.superevents.search', mock_superevents)

    response = client.get(
        url_for('typeahead_superevent_id', superevent_id='MS190235'))

    mock_superevents.assert_not_called()
    assert HTTP_STATUS_CODES[response.status_code] == 'OK'
    assert response.json == []


def test_typeahead_skymap_filename_gracedb_error_404(client, monkeypatch):
    """Test that the typeahead endpoints return an empty list if GraceDB
    returns a 404 error.
    """
    response = Response()
    response.status_code = 404
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_log',
                        Mock(side_effect=HTTPError(response=response)))

    response = client.get(
        url_for('typeahead_skymap_filename', superevent_id='MS190208a'))

    assert HTTP_STATUS_CODES[response.status_code] == 'OK'
    assert response.json == []


def test_typeahead_skymap_filename_gracedb_error_non_404(client, monkeypatch):
    """Test that the typeahead raises an internal error if GraceDB
    returns an error other than 404.
    """
    response = Response()
    response.status_code = 403
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_log',
                        Mock(side_effect=HTTPError(response=response)))

    response = client.get(
        url_for('typeahead_skymap_filename', superevent_id='MS190208a'))

    assert HTTP_STATUS_CODES[response.status_code] == 'Internal Server Error'


@pytest.mark.parametrize('endpoint,tag', [
    ('typeahead_em_bright_filename', 'em_bright'),
    ('typeahead_p_astro_filename', 'p_astro')
])
def test_typeahead_em_bright_and_p_astro(
        endpoint, tag, client, monkeypatch):
    """Test typeahead filtering for em_bright and p_astro files."""
    mock_logs = Mock(return_value=[
        {'file_version': 0,
         'filename': 'foobar.txt',
         'tag_names': [tag]},
        {'file_version': 0,
         'filename': 'bar.json',
         'tag_names': [tag]},
        {'file_version': 0,
         'filename': 'foobar.json',
         'tag_names': [tag]},
        {'file_version': 0,
         'filename': 'foobat.json',
         'tag_names': [tag]},
        {'file_version': 0,
         'filename': 'foobaz.json',
         'tag_names': ['wrong_tag']}])
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_log', mock_logs)

    response = client.get(
        url_for(endpoint, superevent_id='MS190208a', filename='foo'))

    assert HTTP_STATUS_CODES[response.status_code] == 'OK'
    mock_logs.assert_called_once_with('MS190208a')
    assert response.json == ['foobar.json,0', 'foobat.json,0']


def test_download_upload_external_skymap(client, monkeypatch):
    """Test download event from URL and upload to external event."""
    mock_get_upload_external_skymap = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.external_skymaps.get_upload_external_skymap',
        mock_get_upload_external_skymap)
    response = client.post(url_for('download_upload_external_skymap'), data={
        'ext_id': 'E12345', 'skymap_url': 'https://url.fits.gz'
    })
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    mock_get_upload_external_skymap.assert_called_once_with(
            {'graceid': 'E12345', 'search': 'FromURL'},
            skymap_link='https://url.fits.gz')


@pytest.mark.parametrize('search,time_far,space_far', [
    ('GRB', 1e-6, 1e-8),
    ('SNEWS', None, None)
])
def test_apply_raven_labels(search, time_far, space_far,
                            client, monkeypatch):
    """Test assigning RAVEN alert labels and update the preferred
    external event."""
    mock_create_label = Mock()
    mock_gracedb_download = Mock(return_value={
        'temporal_coinc_far': time_far,
        'spatiotemporal_coinc_far': space_far
    })
    mock_update_superevent = Mock()
    mock_get_event = Mock(return_value={
        'graceid': 'E12345',
        'search': search
    })
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.create_label.run',
        mock_create_label)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.download',
        mock_gracedb_download)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.update_superevent',
        mock_update_superevent)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_event.run',
        mock_get_event)
    response = client.post(url_for('apply_raven_labels'), data={
        'superevent_id': 'MS190208a',
        'ext_id': 'E12345',
        'event_id': 'G54321'})
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    mock_create_label.assert_has_calls(
        [call('RAVEN_ALERT', 'MS190208a'),
         call('RAVEN_ALERT', 'E12345'),
         call('RAVEN_ALERT', 'G54321')]
    )
    mock_update_superevent.assert_called_once_with(
        'MS190208a', em_type='E12345',
        time_coinc_far=time_far, space_coinc_far=space_far
    )


def test_apply_raven_labels_no_data(client, monkeypatch):
    """Test assigning a RAVEN alert."""
    response = client.post(url_for('apply_raven_labels'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    assert get_flashed_messages() == [
        'No alert sent. Please fill in all fields.']


def test_send_mock_event(client, monkeypatch):
    """Test creating a mock event."""
    mock_first2years_upload = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.first2years.upload_event.run',
        mock_first2years_upload)
    response = client.post(url_for('send_mock_event'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    mock_first2years_upload.assert_called_once()


def test_send_mock_joint_event(client, monkeypatch):
    """Test creating a joint mock event."""
    mock_first2years_upload = Mock(return_value=100.)
    mock_handle_grb_gcn = Mock()
    monkeypatch.setattr(
        'gwcelery.tasks.first2years.upload_event.run',
        mock_first2years_upload)
    monkeypatch.setattr(
        'gwcelery.tasks.external_triggers.handle_grb_gcn',
        mock_handle_grb_gcn)
    response = client.post(url_for('send_mock_joint_event'))
    assert HTTP_STATUS_CODES[response.status_code] == 'Found'
    mock_first2years_upload.assert_called_once()
    mock_handle_grb_gcn.assert_called_once()



"""Embed an IMAP email client into a Celery worker by :doc:`extending Celery
with bootsteps <celery:userguide/extending>`.
"""
from click import Option

from .bootsteps import Receiver


def install(app):
    """Register the email client subsystem in the application boot steps."""
    app.steps['consumer'] |= {Receiver}
    app.user_options['worker'].add(Option(('--email',),
                                          is_flag=True,
                                          help='Enable email client'))



"""Create and upload external sky maps."""
from astropy import units as u
from astropy.coordinates import ICRS, SkyCoord
import astropy_healpix as ah
from astropy_healpix import HEALPix, pixel_resolution_to_nside
from celery import group
#  import astropy.utils.data
import numpy as np
from ligo.skymap.io import fits
from ligo.skymap.distance import parameters_to_marginal_moments
from ligo.skymap.tool import ligo_skymap_combine
import gcn
import healpy as hp
import io
import lxml.etree
import re
import ssl
import urllib

from ..import app
from . import gracedb
from . import skymaps
from ..util.cmdline import handling_system_exit
from ..util.tempfile import NamedTemporaryFile
from ..import _version


COMBINED_SKYMAP_FILENAME_MULTIORDER = 'combined-ext.multiorder.fits'
"""Filename of combined sky map in a multiordered format"""

COMBINED_SKYMAP_FILENAME_FLAT = 'combined-ext.fits.gz'
"""Filename of combined sky map in a flattened format"""

COMBINED_SKYMAP_FILENAME_PNG = 'combined-ext.png'
"""Filename of combined sky map plot"""


@app.task(shared=False,
          queue='exttrig')
def create_combined_skymap(se_id, ext_id, preferred_event=None):
    """Creates and uploads the combined LVK-external skymap, uploading to the
    external trigger GraceDB page. The filename used for the combined sky map
    will be 'combined-ext.multiorder.fits' if the GW sky map is multi-ordered
    or 'combined-ext.fits.gz' if the GW sky map is not. Will use the GW sky map
    in from the preferred event if given.

    Parameters
    ----------
    se_id : str
        Superevent GraceDB ID
    ext_id: str
        External event GraceDB ID
    preferred_event: str
        Preferred event GraceDB ID. If given, use sky map from preferred event

    """
    # gw_id is either superevent or preferred event graceid
    gw_id = preferred_event if preferred_event else se_id
    gw_skymap_filename = get_skymap_filename(gw_id, is_gw=True)
    ext_skymap_filename = get_skymap_filename(ext_id, is_gw=False)
    # Determine whether GW sky map is multiordered or flat
    gw_moc = '.multiorder.fits' in gw_skymap_filename

    new_filename = \
        (COMBINED_SKYMAP_FILENAME_MULTIORDER if gw_moc else
         COMBINED_SKYMAP_FILENAME_FLAT)

    message = \
        ('Combined LVK-external sky map using {0} and {1}, with {2} and '
         '{3}'.format(gw_id, ext_id, gw_skymap_filename, ext_skymap_filename))
    message_png = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=ext_id,
            filename=new_filename)

    (
        _download_skymaps.si(
            gw_skymap_filename, ext_skymap_filename, gw_id, ext_id
        )
        |
        combine_skymaps.s(gw_moc=gw_moc)
        |
        group(
            gracedb.upload.s(
                new_filename, ext_id, message, ['sky_loc', 'ext_coinc']
            )
            |
            gracedb.create_label.si('COMBINEDSKYMAP_READY', ext_id),

            skymaps.plot_allsky.s()
            |
            gracedb.upload.s(COMBINED_SKYMAP_FILENAME_PNG, ext_id,
                             message_png, ['sky_loc', 'ext_coinc'])
        )
    ).delay()


@app.task(autoretry_for=(ValueError,), retry_backoff=10,
          queue='exttrig',
          retry_backoff_max=600)
def get_skymap_filename(graceid, is_gw):
    """Get the skymap fits filename.

    If not available, will try again 10 seconds later, then 20, then 40, etc.
    until up to 10 minutes after initial attempt.

    Parameters
    ----------
    graceid : str
        GraceDB ID
    is_gw: bool
        If True, uses method for superevent or preferred event. Otherwise uses
        method for external event.

    """
    gracedb_log = gracedb.get_log(graceid)
    if is_gw:
        # Try first to get a multiordered sky map
        for message in reversed(gracedb_log):
            filename = message['filename']
            v = message['file_version']
            fv = '{},{}'.format(filename, v)
            if filename.endswith('.multiorder.fits') and \
                    "combined-ext." not in filename:
                return fv
        # Try next to get a flattened sky map
        for message in reversed(gracedb_log):
            filename = message['filename']
            v = message['file_version']
            fv = '{},{}'.format(filename, v)
            if filename.endswith('.fits.gz') and \
                    "combined-ext." not in filename:
                return fv
    else:
        for message in reversed(gracedb_log):
            filename = message['filename']
            v = message['file_version']
            fv = '{},{}'.format(filename, v)
            if (filename.endswith('.fits') or filename.endswith('.fit') or
                    filename.endswith('.fits.gz')) and \
                    "combined-ext." not in filename:
                return fv
    raise ValueError('No skymap available for {0} yet.'.format(graceid))


@app.task(shared=False, queue='exttrig')
def _download_skymaps(gw_filename, ext_filename, gw_id, ext_id):
    """Download both superevent and external sky map to be combined."""
    gw_skymap = gracedb.download(gw_filename, gw_id)
    ext_skymap = gracedb.download(ext_filename, ext_id)
    return gw_skymap, ext_skymap


def combine_skymaps_moc_flat(gw_sky, ext_sky, ext_header):
    """This function combines a multiordered (MOC) GW sky map with a flattened
    external one by reweighting the MOC sky map using the values of the
    flattened one.

    Header info is generally inherited from the GW sky map or recalculated
    using the combined sky map values.
    """
    #  Find ra/dec of each GW pixel
    level, ipix = ah.uniq_to_level_ipix(gw_sky["UNIQ"])
    nsides = ah.level_to_nside(level)
    areas = ah.nside_to_pixel_area(nsides)
    ra_gw, dec_gw = ah.healpix_to_lonlat(ipix, nsides, order='nested')
    #  Find corresponding external sky map indicies
    ext_nside = ah.npix_to_nside(len(ext_sky))
    ext_ind = ah.lonlat_to_healpix(
        ra_gw, dec_gw, ext_nside,
        order='nested' if ext_header['nest'] else 'ring')
    #  Reweight GW prob density by external sky map probabilities
    gw_sky['PROBDENSITY'] *= ext_sky[ext_ind]
    gw_sky['PROBDENSITY'] /= \
        np.sum(gw_sky['PROBDENSITY'] * areas).value
    #  Modify GW sky map with new data, ensuring they exist first
    if 'DISTMU' in gw_sky.keys() and 'DISTSIGMA' in gw_sky.keys():
        distmean, diststd = parameters_to_marginal_moments(
            gw_sky['PROBDENSITY'] * areas.value,
            gw_sky['DISTMU'], gw_sky['DISTSIGMA'])
        gw_sky.meta['distmean'], gw_sky.meta['diststd'] = distmean, diststd
    if 'instruments' not in ext_header:
        ext_header.update({'instruments': {'external instrument'}})
    if 'instruments' in gw_sky.meta:
        gw_sky.meta['instruments'].update(ext_header['instruments'])
    if 'HISTORY' in gw_sky.meta:
        ext_instrument = list(ext_header['instruments'])[0]
        gw_sky.meta['HISTORY'].extend([
            '', 'The values were reweighted by using data from {0}{1}'.format(
                ('an ' if ext_instrument == 'external instrument'
                 else ''),
                ext_instrument)])
    return gw_sky


@app.task(shared=False, queue='exttrig')
def combine_skymaps(skymapsbytes, gw_moc=True):
    """This task combines the two input skymaps, in this case the external
    trigger skymap and the LVK skymap and writes to a temporary output file. It
    then returns the contents of the file as a byte array.

    There are separate methods in case the GW sky map is multiordered (we just
    reweight using the external sky map) or flattened (use standard
    ligo.skymap.combine method).
    """
    gw_skymap_bytes, ext_skymap_bytes = skymapsbytes
    suffix = ".fits" if gw_moc else ".fits.gz"
    with NamedTemporaryFile(mode='rb', suffix=suffix) as combinedskymap, \
            NamedTemporaryFile(content=gw_skymap_bytes) as gw_skymap_file, \
            NamedTemporaryFile(content=ext_skymap_bytes) as ext_skymap_file, \
            handling_system_exit():

        # If GW sky map is multiordered, use reweighting method
        if gw_moc:
            #  FIXME: Use method that regrids the combined sky map e.g. mhealpy
            #  once this method is quicker and preserves the header
            #  Load sky maps
            gw_skymap = fits.read_sky_map(gw_skymap_file.name, moc=True)
            ext_skymap, ext_header = fits.read_sky_map(ext_skymap_file.name,
                                                       moc=False)
            #  Create and write combined sky map
            combined_skymap = combine_skymaps_moc_flat(gw_skymap, ext_skymap,
                                                       ext_header)
            fits.write_sky_map(combinedskymap.name, combined_skymap, moc=True)
        # If GW sky map is flattened, use older method
        else:
            ligo_skymap_combine.main([gw_skymap_file.name,
                                      ext_skymap_file.name,
                                      combinedskymap.name])
        #  FIXME: Add method for MOC-MOC if there is a switch to MOC external
        #  skymaps
        return combinedskymap.read()


@app.task(shared=False)
def external_trigger(graceid):
    """Returns the associated external trigger GraceDB ID."""
    em_events = gracedb.get_superevent(graceid)['em_events']
    if len(em_events):
        for exttrig in em_events:
            if gracedb.get_event(exttrig)['search'] == 'GRB':
                return exttrig
    raise ValueError('No associated GRB EM event(s) for {0}.'.format(graceid))


@app.task(shared=False, queue='exttrig')
def external_trigger_heasarc(external_id):
    """Returns the HEASARC fits file link."""
    gracedb_log = gracedb.get_log(external_id)
    for message in gracedb_log:
        if 'Original Data' in message['comment']:
            filename = message['filename']
            xmlfile = gracedb.download(urllib.parse.quote(filename),
                                       external_id)
            root = lxml.etree.fromstring(xmlfile)
            heasarc_url = root.find('./What/Param[@name="LightCurve_URL"]'
                                    ).attrib['value']
            return re.sub(r'quicklook(.*)', 'current/', heasarc_url)
    raise ValueError('Not able to retrieve HEASARC link for {0}.'.format(
        external_id))


@app.task(autoretry_for=(urllib.error.HTTPError,), retry_backoff=10,
          queue='exttrig',
          retry_backoff_max=600)
def get_external_skymap(link, search):
    """Download the Fermi sky map fits file and return the contents as a byte
    array. If GRB, will construct a HEASARC url, while if SubGRB, will use the
    link directly.

    If not available, will try again 10 seconds later, then 20, then 40, etc.
    until up to 10 minutes after initial attempt.
    """
    if search == 'GRB':
        # if Fermi GRB, determine final HEASARC link
        trigger_id = re.sub(r'.*\/(\D+?)(\d+)(\D+)\/.*', r'\2', link)
        skymap_name = 'glg_healpix_all_bn{0}_v00.fit'.format(trigger_id)
        skymap_link = link + skymap_name
    elif search in {'SubGRB', 'FromURL'}:
        skymap_link = link
    #  FIXME: Under Anaconda on the LIGO Caltech computing cluster, Python
    #  (and curl, for that matter) fail to negotiate TLSv1.2 with
    #  heasarc.gsfc.nasa.gov
    context = ssl.create_default_context()
    context.options |= ssl.OP_NO_TLSv1_3
    #  return astropy.utils.data.get_file_contents(
    #      (skymap_link), encoding='binary', cache=False)
    return urllib.request.urlopen(skymap_link, context=context).read()


@app.task(autoretry_for=(urllib.error.HTTPError, urllib.error.URLError,),
          queue='exttrig',
          retry_backoff=10, retry_backoff_max=1200)
def get_upload_external_skymap(event, skymap_link=None):
    """If a Fermi sky map is not uploaded yet, tries to download one and upload
    to external event. If sky map is not available, passes so that this can be
    re-run the next time an update GCN notice is received. If GRB, will
    construct a HEASARC url, while if SubGRB, will use the link directly.
    If SubGRB or FromURL, downloads a skymap using the provided URL rather
    than construct one.
    """
    graceid = event['graceid']
    search = event['search']

    if search == 'GRB':
        external_skymap_canvas = (
            external_trigger_heasarc.si(graceid)
            |
            get_external_skymap.s(search)
        )
    elif search in {'SubGRB', 'FromURL'}:
        external_skymap_canvas = get_external_skymap.si(skymap_link, search)

    skymap_filename = \
        ('external_from_url' if search == 'FromURL'
         else 'glg_healpix_all_bn_v00')

    fits_message = \
        ('Downloaded from {}.'.format(skymap_link) if search == 'FromURL'
         else 'Official sky map from Fermi analysis.')
    png_message = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=graceid, filename=skymap_filename + '.fits')

    (
        external_skymap_canvas
        |
        group(
            gracedb.upload.s(
                skymap_filename + '.fits',
                graceid,
                fits_message,
                ['sky_loc']),

            skymaps.plot_allsky.s()
            |
            gracedb.upload.s(skymap_filename + '.png',
                             graceid,
                             png_message,
                             ['sky_loc'])
        )
        |
        gracedb.create_label.si('EXT_SKYMAP_READY', graceid)
    ).delay()


def create_external_skymap(ra, dec, error, pipeline, notice_type=111):
    """Create a sky map, either a gaussian or a single
    pixel sky map, given an RA, dec, and error radius.

    If from Fermi, convolves the sky map with both a core and
    tail Gaussian and then sums these to account for systematic
    effects as measured in :doi:`10.1088/0067-0049/216/2/32`

    If from Swift, converts the error radius from that containing 90% of the
    credible region to ~68% (see description of Swift error
    here:`https://gcn.gsfc.nasa.gov/swift.html#tc7`)

    Parameters
    ----------
    ra : float
        right ascension in deg
    dec: float
        declination in deg
    error: float
        error radius in deg

    Returns
    -------
    skymap : numpy array
        sky map array

    """
    max_nside = 2048
    if error:
        # Correct 90% containment to 1-sigma for Swift
        if pipeline == 'Swift':
            error /= np.sqrt(-2 * np.log1p(-.9))
        error_radius = error * u.deg
        # Use minimum nside of 256
        nside = max(pixel_resolution_to_nside(error_radius, round='up'), 256)
    else:
        nside = np.inf
    if nside >= max_nside:
        nside = max_nside

        #  Find the one pixel the event can localized to
        hpx = HEALPix(nside, 'ring', frame=ICRS())
        skymap = np.zeros(hpx.npix)
        ind = hpx.lonlat_to_healpix(ra * u.deg, dec * u.deg)
        skymap[ind] = 1.
    else:
        #  If larger error, create gaussian sky map
        hpx = HEALPix(nside, 'ring', frame=ICRS())
        ipix = np.arange(hpx.npix)

        #  Evaluate Gaussian.
        center = SkyCoord(ra * u.deg, dec * u.deg)
        distance = hpx.healpix_to_skycoord(ipix).separation(center)
        skymap = np.exp(-0.5 * np.square(distance / error_radius).to_value(
            u.dimensionless_unscaled))
        skymap /= skymap.sum()
    if pipeline == 'Fermi':
        # Correct for Fermi systematics based on recommendations from GBM team
        # Convolve with both a narrow core and wide tail Gaussian with error
        # radius determined by the scales respectively, each comprising a
        # fraction determined by the weights respectively
        if notice_type == gcn.NoticeType.FERMI_GBM_FLT_POS:
            # Flight notice
            # Values from first row of Table 7
            weights = [0.897, 0.103]
            scales = [7.52, 55.6]
        elif notice_type == gcn.NoticeType.FERMI_GBM_GND_POS:
            # Ground notice
            # Values from first row of Table 3
            weights = [0.804, 0.196]
            scales = [3.72, 13.7]
        elif notice_type == gcn.NoticeType.FERMI_GBM_FIN_POS:
            # Final notice
            # Values from second row of Table 3
            weights = [0.900, 0.100]
            scales = [3.71, 14.3]
        else:
            raise AssertionError(
                'Need to provide a supported Fermi notice type')
        skymap = sum(
            weight * hp.sphtfunc.smoothing(skymap, sigma=np.radians(scale))
            for weight, scale in zip(weights, scales))

    # Renormalize due to possible lack of precision
    # Enforce the skymap to be non-negative
    return np.abs(skymap) / np.abs(skymap).sum()


def write_to_fits(skymap, event, notice_type, notice_date):
    """Write external sky map fits file, populating the
    header with relevant info.

    Parameters
    ----------
    skymap : numpy array
        sky map array
    event : dict
        Dictionary of Swift external event

    Returns
    -------
    skymap fits : bytes array
        bytes array of sky map

    """
    notice_type_dict = {
        '53': 'INTEGRAL_WAKEUP',
        '54': 'INTEGRAL_REFINED',
        '55': 'INTEGRAL_OFFLINE',
        '60': 'SWIFT_BAT_GRB_ALERT',
        '61': 'SWIFT_BAT_GRB_POSITION',
        '105': 'AGILE_MCAL_ALERT',
        '110': 'FERMI_GBM_ALERT',
        '111': 'FERMI_GBM_FLT_POS',
        '112': 'FERMI_GBM_GND_POS',
        '115': 'FERMI_GBM_FINAL_POS',
        '131': 'FERMI_GBM_SUBTHRESHOLD'}

    if notice_type is None:
        msgtype = event['pipeline'] + '_LVK_TARGETED_SEARCH'
    else:
        msgtype = notice_type_dict[str(notice_type)]

    gcn_id = event['extra_attributes']['GRB']['trigger_id']
    with NamedTemporaryFile(suffix='.fits.gz') as f:
        fits.write_sky_map(f.name, skymap,
                           objid=gcn_id,
                           url=event['links']['self'],
                           instruments=event['pipeline'],
                           gps_time=event['gpstime'],
                           msgtype=msgtype,
                           msgdate=notice_date,
                           creator='gwcelery',
                           origin='LIGO-VIRGO-KAGRA',
                           vcs_version=_version.get_versions()['version'],
                           history='file only for internal use')
        with open(f.name, 'rb') as file:
            return file.read()


@app.task(shared=False, queue='exttrig')
def create_upload_external_skymap(event, notice_type, notice_date):
    """Create and upload external sky map using
    RA, dec, and error radius information.

    Parameters
    ----------
    event : dict
        Dictionary of Swift external event

    """
    graceid = event['graceid']
    skymap_filename = event['pipeline'].lower() + '_skymap.fits.gz'

    ra = event['extra_attributes']['GRB']['ra']
    dec = event['extra_attributes']['GRB']['dec']
    error = event['extra_attributes']['GRB']['error_radius']
    pipeline = event['pipeline']

    if not (ra or dec or error):
        # Don't create sky map if notice only contains zeros, lacking info
        return
    skymap = create_external_skymap(ra, dec, error, pipeline, notice_type)

    skymap_data = write_to_fits(skymap, event, notice_type, notice_date)

    message = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=graceid, filename=skymap_filename)

    (
        gracedb.upload.si(skymap_data,
                          skymap_filename,
                          graceid,
                          'Sky map created from GCN RA, dec, and error.',
                          ['sky_loc'])
        |
        skymaps.plot_allsky.si(skymap_data, ra=ra, dec=dec)
        |
        gracedb.upload.s(event['pipeline'].lower() + '_skymap.png',
                         graceid,
                         message,
                         ['sky_loc'])
        |
        gracedb.create_label.si('EXT_SKYMAP_READY', graceid)
    ).delay()


@app.task(shared=False)
def plot_overlap_integral(coinc_far_dict, superevent, ext_event,
                          var_label=r"\mathcal{I}_{\Omega}"):
    """Plot and upload visualization of the sky map overlap integral computed
    by ligo.search.overlap_integral.

    Parameters
    ----------
    coinc_far_dict : dict
        Dictionary containing coincidence false alarm rate results from
        RAVEN
    superevent : dict
        superevent dictionary
    ext_event : dict
        external event dictionary
    var_label : str
        The variable symbol used in plotting

    """
    if coinc_far_dict.get('skymap_overlap') is None:
        return
    if superevent['em_type'] != ext_event['graceid'] and \
            'RAVEN_ALERT' in superevent['labels']:
        return

    superevent_id = superevent['superevent_id']
    ext_id = ext_event['graceid']

    log_overlap = np.log(coinc_far_dict['skymap_overlap'])
    logI_string = np.format_float_positional(log_overlap, 1, trim='0',
                                             sign=True)
    # Create plot
    fig, _ = skymaps.plot_bayes_factor(
        log_overlap, values=(1, 3, 5), xlim=7, var_label=var_label,
        title=(r'Sky Map Overlap between %s and %s [$\ln\,%s = %s$]' %
               (superevent_id, ext_id, var_label, logI_string)))
    # Convert to bytes
    outfile = io.BytesIO()
    fig.savefig(outfile, format='png')
    # Upload file
    gracedb.upload.si(
        outfile.getvalue(),
        'overlap_integral.png',
        superevent_id,
        message='Sky map overlap integral between {0} and {1}'.format(
            superevent_id, ext_id),
        tags=['ext_coinc']
    ).delay()



import json

from astropy import time
from celery import group
from celery.utils.log import get_logger
import numpy as np

from ..import app
from .core import DispatchHandler
from . import gracedb
from ..kafka.signals import kafka_record_consumed

log = get_logger(__name__)


class _KafkaDispatchHandler(DispatchHandler):

    def process_args(self, name, record):
        return name, (record,), {}


handler = _KafkaDispatchHandler()
r"""Function decorator to register a handler callback for specified Kafka URLs.
The decorated function is turned into a Celery task, which will be
automatically called whenever a message is received from a matching URL.

Parameters
----------
\*keys
    List of keys from :obj:`gwcelery.conf.kafka_consumer_config`
    associated with Kafka topics to listen to messages to.
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    # Assumes kafka_consumer_config dictionary has 'fermi_swift' key
    @alerts.handler('fermi_swift')
    def handle_swift(record):
        # record is a dict that contains the contents of the message
        # do work here...
"""


@kafka_record_consumed.connect
def _on_kafka_record_consumed(name, record, **kwargs):
    handler.dispatch(name, record)


def _create_base_alert_dict(classification, superevent, alert_type):
    '''Create the base of the alert dictionary, with all contents except the
    skymap and the external coinc information.'''
    # NOTE Everything that comes through this code path will be marked as
    # public. However, MDC events with this flag are not made public on
    # GraceDB-playground and GraceDB-test.
    # Re time_created: Dont need better than second precision for alert times

    # NOTE less-significant alerts have alert_type as PRELIMINARY
    alert_type_kafka = 'preliminary' if alert_type == 'less-significant' \
        else alert_type

    alert_dict = {
        'alert_type': alert_type_kafka.upper(),
        'time_created': time.Time.now().utc.isot.split('.')[0] + 'Z',
        'superevent_id': superevent['superevent_id'],
        'urls': {'gracedb': superevent['links']['self'].replace('api/', '') +
                 'view/'},
        'event': None,
        'external_coinc': None
    }

    if alert_type == 'retraction':
        return alert_dict

    if classification and classification[0] is not None:
        properties = json.loads(classification[0])
    else:
        properties = {}

    if classification and classification[1] is not None:
        classification = json.loads(classification[1])
    else:
        classification = {}

    duration = None
    central_frequency = None

    if superevent['preferred_event_data']['group'] == 'Burst':
        if superevent['preferred_event_data']['pipeline'].lower() == 'cwb':
            duration = \
                superevent['preferred_event_data']['extra_attributes'].get(
                    'MultiBurst', {}).get('duration', None)
            central_frequency = \
                superevent['preferred_event_data']['extra_attributes'].get(
                    'MultiBurst', {}).get('central_freq', None)
        elif superevent['preferred_event_data']['pipeline'].lower() == 'mly':
            duration = \
                superevent['preferred_event_data']['extra_attributes'].get(
                    'MLyBurst', {}).get('duration', None)
            central_frequency = \
                superevent['preferred_event_data']['extra_attributes'].get(
                    'MLyBurst', {}).get('central_freq', None)
        elif superevent['preferred_event_data']['pipeline'].lower() == 'olib':
            quality_mean = \
                superevent['preferred_event_data']['extra_attributes'].get(
                     'LalInferenceBurst', {}).get('quality_mean', None)
            frequency_mean = \
                superevent['preferred_event_data']['extra_attributes'].get(
                     'LalInferenceBurst', {}).get('frequency_mean', None)
            central_frequency = \
                superevent['preferred_event_data']['extra_attributes'].get(
                     'LalInferenceBurst', {}).get('frequency_mean', None)
            duration = quality_mean / (2 * np.pi * frequency_mean)
        else:
            raise NotImplementedError(
                'Duration and central_frequency not implemented for Burst '
                'pipeline {}'.format(
                    superevent['preferred_event_data']['pipeline'].lower()
                )
            )

    alert_dict['event'] = {
        # set 'significant' field based on
        # https://dcc.ligo.org/LIGO-G2300151/public
        'significant': False if alert_type == 'less-significant' else True,
        'time': time.Time(superevent['t_0'], format='gps').utc.isot + 'Z',
        'far': superevent['far'],
        'instruments': sorted(
            superevent['preferred_event_data']['instruments'].split(',')
        ),
        'group': superevent['preferred_event_data']['group'],
        'pipeline': superevent['preferred_event_data']['pipeline'],
        'search': superevent['preferred_event_data']['search'],
        'properties': properties,
        'classification': classification,
        'duration': duration,
        'central_frequency': central_frequency
    }

    return alert_dict


@gracedb.task(shared=False)
def _add_external_coinc_to_alert(alert_dict, superevent,
                                 combined_skymap_filename):
    external_event = gracedb.get_event(superevent['em_type'])
    if combined_skymap_filename:
        combined_skymap = gracedb.download(combined_skymap_filename,
                                           superevent['superevent_id'])
    else:
        combined_skymap = None
    alert_dict['external_coinc'] = {
        'gcn_notice_id':
            int(external_event['extra_attributes']['GRB']['trigger_id']),
        'ivorn': external_event['extra_attributes']['GRB']['ivorn'],
        'observatory': external_event['pipeline'],
        'search': external_event['search'],
        'time_difference': round(superevent['t_0']
                                 - external_event['gpstime'], 2),
        'time_coincidence_far': superevent['time_coinc_far'],
        'time_sky_position_coincidence_far': superevent['space_coinc_far']
    }

    return alert_dict, combined_skymap


@app.task(bind=True, shared=False, queue='kafka', ignore_result=True)
def _upload_notice(self, payload, brokerhost, superevent_id):
    '''
    Upload serialized alert notice to GraceDB
    '''
    config = self.app.conf['kafka_alert_config'][brokerhost]
    kafka_writer = self.app.conf['kafka_streams'][brokerhost]

    # FIXME Drop get_payload_content method once
    # https://github.com/scimma/hop-client/pull/190 is merged
    alert_dict = kafka_writer.get_payload_content(payload)
    message = 'Kafka alert notice sent to {}'.format(config['url'])

    filename = '{}-{}.{}'.format(
        alert_dict['superevent_id'],
        alert_dict['alert_type'].lower(),
        config['suffix']
    )

    gracedb.upload.delay(payload.serialize()['content'], filename,
                         superevent_id, message, tags=['public', 'em_follow'])


@app.task(bind=True, queue='kafka', shared=False)
def _send(self, alert_dict, skymap, brokerhost, combined_skymap=None):
    """Write the alert to the Kafka topic"""
    # Copy the alert dictionary so we dont modify the original
    payload_dict = alert_dict.copy()
    # Add skymap to alert_dict
    config = self.app.conf['kafka_alert_config'][brokerhost]
    if alert_dict['event'] is not None:
        # dict.copy is a shallow copy, so need to copy event dict as well since
        # we plan to modify it
        payload_dict['event'] = alert_dict['event'].copy()

        # Encode the skymap
        encoder = config['skymap_encoder']
        payload_dict['event']['skymap'] = encoder(skymap)

        if combined_skymap:
            payload_dict['external_coinc']['combined_skymap'] = \
                encoder(combined_skymap)

    # Write to kafka topic
    serialization_model = \
        self.app.conf['kafka_streams'][brokerhost].serialization_model
    payload = serialization_model(payload_dict)
    self.app.conf['kafka_streams'][brokerhost].write(payload)

    return payload


@app.task(bind=True, queue='kafka', shared=False)
def _send_with_combined(self, alert_dict_combined_skymap, skymap, brokerhost):
    alert_dict, combined_skymap = alert_dict_combined_skymap
    return _send(alert_dict, skymap, brokerhost,
                 combined_skymap=combined_skymap)


@app.task(bind=True, ignore_result=True, queue='kafka', shared=False)
def send(self, skymap_and_classification, superevent, alert_type,
         raven_coinc=False, combined_skymap_filename=None):
    """Send an public alert to all currently connected kafka brokers.

    Parameters
    ----------
    skymap_and_classification : tuple, None
        The filecontents of the skymap followed by a collection of JSON
        strings. The former generated by
        :meth:`gwcelery.tasks.gracedb.download`, the latter generated by
        :meth:`gwcelery.tasks.em_bright.classifier` and
        :meth:`gwcelery.tasks.p_astro.compute_p_astro` or content of
        ``{gstlal,mbta}.p_astro.json`` uploaded by {gstlal,mbta} respectively.
        Can also be None.
    superevent : dict
        The superevent dictionary, typically obtained from an IGWN Alert or
        from querying GraceDB.
    alert_type : str
        The alert type. Either of {`less-significant`, `earlywarning`,
        `preliminary`, `initial`, `update`}.
    raven_coinc: bool
        Is there a coincident external event processed by RAVEN?
    combined_skymap_filename : str
        Combined skymap filename. Default None.

    Notes
    -----
    The `alert_type` value is used to set the `significant` field in the
    alert dictionary.
    """

    if skymap_and_classification is not None:
        skymap, *classification = skymap_and_classification
    else:
        skymap = None
        classification = None

    alert_dict = _create_base_alert_dict(
        classification,
        superevent,
        alert_type
    )

    if raven_coinc and alert_type != 'retraction':
        canvas = (
            _add_external_coinc_to_alert.si(
                alert_dict,
                superevent,
                combined_skymap_filename
            )
            |
            group(
                (
                    _send_with_combined.s(skymap, brokerhost)
                    |
                    _upload_notice.s(brokerhost, superevent['superevent_id'])
                ) for brokerhost in self.app.conf['kafka_streams'].keys()
            )
        )
    else:
        canvas = (
            group(
                (
                    _send.s(alert_dict, skymap, brokerhost)
                    |
                    _upload_notice.s(brokerhost, superevent['superevent_id'])
                ) for brokerhost in self.app.conf['kafka_streams'].keys()
            )
        )

    canvas.apply_async()


@app.task(shared=False)
def _create_skymap_classification_tuple(skymap, classification):
    return (skymap, *classification)


@app.task(shared=False, ignore_result=True)
def download_skymap_and_send_alert(classification, superevent, alert_type,
                                   skymap_filename=None, raven_coinc=False,
                                   combined_skymap_filename=None):
    """Wrapper for send function when caller has not already downloaded the
    skymap.

    Parameters
    ----------
    classification : tuple, None
        A collection of JSON strings, generated by
        :meth:`gwcelery.tasks.em_bright.classifier` and
        :meth:`gwcelery.tasks.p_astro.compute_p_astro` or
        content of ``{gstlal,mbta}.p_astro.json`` uploaded by {gstlal,mbta}
        respectively; or None
    superevent : dict
        The superevent dictionary, typically obtained from an IGWN Alert or
        from querying GraceDB.
    alert_type : {'earlywarning', 'preliminary', 'initial', 'update'}
        The alert type.
    skymap_filename : string
        The skymap filename.
    raven_coinc: bool
        Is there a coincident external event processed by RAVEN?
    combined_skymap_filename : str
        The combined skymap filename. Default None
    """

    if skymap_filename is not None and alert_type != 'retraction':
        canvas = (
            gracedb.download.si(
                skymap_filename,
                superevent['superevent_id']
            )
            |
            _create_skymap_classification_tuple.s(classification)
            |
            send.s(superevent, alert_type, raven_coinc=raven_coinc,
                   combined_skymap_filename=combined_skymap_filename)
        )
    else:
        canvas = send.s(
            (None, classification),
            superevent,
            alert_type,
            raven_coinc=raven_coinc,
            combined_skymap_filename=combined_skymap_filename
        )

    canvas.apply_async()



"""Detect if we are running under Sphinx."""
__all__ = ('SPHINX',)


try:
    # This global builtin variable is set in doc/conf.py.
    # See https://stackoverflow.com/a/65147676/167694
    SPHINX = __sphinx_build__
except NameError:
    SPHINX = False



"""Definitions of custom :doc:`Celery signals <celery:userguide/signals>`
related to IGWN alerts.

These signals allow us to keep the IGWN alert broker code decoupled from any
GCN-specific logic. Notably, it allows us to keep all of the details of
the GCN-specific "Notice Type" concept out of :mod:`gwcelery.voevent`.
"""
from celery.utils.dispatch import Signal

igwn_alert_received = Signal(
    name='igwn_alert_received', providing_args=('topic', 'payload'))
"""Fired whenever a IGWN alert is received.

Parameters
----------
topic : str
    The igwn alert topic
payload : dict
    Alert dictionary
"""



from importlib import resources
import json
from unittest.mock import Mock, patch

import pytest

from .. import app
from ..tasks import alerts
from . import data


@patch('gwcelery.tasks.alerts._upload_notice.run')
@patch('gwcelery.tasks.gracedb.download._orig_run',
       return_value=resources.read_binary(
           data,
           'MS220722v_bayestar.multiorder.fits'
       ))  # Using only bayestar sky map because content doesnt matter
def test_not_implemented_burst_pipeline(mock_download, mock_upload,
                                        monkeypatch):
    """Confirm that using an event from a  Burst pipeline that isn't
    cwb/mly/olib to generate an alert will raise a NotImplementedError.
    """

    # Replace the stream object with a mock object with _validiate_alert as its
    # write attribute
    mock_stream = Mock()
    mock_stream_write = Mock()
    mock_stream.write = mock_stream_write
    monkeypatch.setitem(app.conf, 'kafka_streams', {'scimma': mock_stream})

    # Load superevent dictionary from cwb, change pipeline name
    superevent_dict = json.loads(resources.read_binary(data, 'S230413b.json'))
    superevent_dict['preferred_event_data']['pipeline'] = 'RealFakeDoors'
    skymap = 'MS220722v_bayestar.multiorder.fits'

    with pytest.raises(NotImplementedError):
        # Change the superevent_id field of the superevent to None because the
        # schema does not allow None, so we know this should make the
        # validation fail
        alerts.download_skymap_and_send_alert([None, None], superevent_dict,
                                              'initial', skymap)

    mock_upload.assert_not_called()
    mock_stream_write.assert_not_called()



"""Application configuration for ``gracedb-test.ligo.org``.

Inherits all settings from :mod:`gwcelery.conf.playground`, with the exceptions
below.
"""

from base64 import b64encode

from . import *  # noqa: F401, F403

igwn_alert_group = 'gracedb-test'
"""IGWN alert group."""

gracedb_host = 'gracedb-test.ligo.org'
"""GraceDB host."""

kafka_alert_config = {
    'scimma': {'url': 'kafka://kafka.scimma.org/igwn.gwalert-test',
               'suffix': 'avro', 'skymap_encoder': lambda _: _},
    'gcn': {'url': 'kafka://kafka.dev.gcn.nasa.gov/igwn.gwalert',
            'suffix': 'json', 'skymap_encoder': lambda b:
            b64encode(b).decode('utf-8')}
}
"""Kafka broker configuration details"""

llhoft_glob = '/dev/shm/kafka/{detector}/*.gwf'
"""File glob for low-latency h(t) frames."""

low_latency_frame_types = {'H1': 'H1_llhoft',
                           'L1': 'L1_llhoft',
                           'V1': 'V1_llhoft'}
"""Types of frames used in Parameter Estimation (see
:mod:`gwcelery.tasks.inference`) and in cache creation for detchar
checks (see :mod:`gwcelery.tasks.detchar`).
"""

high_latency_frame_types = {'H1': 'H1_HOFT_C00',
                            'L1': 'L1_HOFT_C00',
                            'V1': 'V1Online'}
"""Types of high latency frames used in Parameter Estimation
(see :mod:`gwcelery.tasks.inference`) and in cache creation for detchar
checks (see :mod:`gwcelery.tasks.detchar`).
"""

strain_channel_names = {'H1': 'H1:GDS-CALIB_STRAIN_CLEAN',
                        'L1': 'L1:GDS-CALIB_STRAIN_CLEAN',
                        'V1': 'V1:Hrec_hoft_16384Hz'}
"""Names of h(t) channels used in Parameter Estimation (see
:mod:`gwcelery.tasks.inference`)"""

sentry_environment = 'test'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`_ in Sentry log
messages."""

mock_events_simulate_multiple_uploads = True
"""If True, then upload each mock event several times in rapid succession with
random jitter in order to simulate multiple pipeline uploads."""



from importlib import resources
import logging
from unittest.mock import patch

from comet.utility.xml import xml_document
import lxml.etree
import pytest

from ..tasks import gcn
from . import data

logging.basicConfig(level=logging.INFO)


def fake_gcn(notice_type):
    # Check the real GCN notice, which is valid.
    payload = resources.read_binary(data, 'G298048-1-Initial.xml')
    root = lxml.etree.fromstring(payload)
    notice_type = str(int(notice_type))
    root.find(".//Param[@name='Packet_Type']").attrib['value'] = notice_type
    return xml_document(root)


def test_unrecognized_notice_type(caplog):
    """Test handling an unrecognized (enum not defined) notice type."""
    caplog.set_level(logging.WARNING)
    gcn.handler.dispatch(fake_gcn(10000))
    record, = caplog.records
    assert record.message == 'ignoring unrecognized key: 10000'


def test_unregistered_notice_type(caplog):
    """Test handling an unregistered notice type."""
    caplog.set_level(logging.WARNING)
    gcn.handler.dispatch(fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
    record, = caplog.records
    assert record.message == ('ignoring unrecognized key: '
                              '<NoticeType.SWIFT_UVOT_POS_NACK: 89>')


@pytest.fixture
def reset_handlers():
    old_handler = dict(gcn.handler)
    gcn.handler.clear()
    yield
    gcn.handler.update(old_handler)


def test_registered_notice_type(reset_handlers):
    @gcn.handler(gcn.NoticeType.AGILE_POINTDIR, gcn.NoticeType.AGILE_TRANS)
    def agile_handler(payload):
        pass

    with patch.object(agile_handler, 'run') as mock_run:
        gcn.handler.dispatch(fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
        mock_run.assert_not_called()
        gcn.handler.dispatch(fake_gcn(gcn.NoticeType.AGILE_POINTDIR))
        mock_run.assert_called_once()



"""Annotations for sky maps."""
import io
import os
import tempfile

from astropy.io import fits
from astropy import table
from celery import group
from celery.exceptions import Ignore
from ligo.skymap.tool import ligo_skymap_flatten
from ligo.skymap.tool import ligo_skymap_unflatten
from ligo.skymap.tool import ligo_skymap_from_samples
from ligo.skymap.tool import ligo_skymap_plot
from ligo.skymap.tool import ligo_skymap_plot_volume
from matplotlib import pyplot as plt
import numpy as np

from . import external_skymaps
from . import gracedb
from . import igwn_alert
from ..import app
from ..jinja import env
from ..util.cmdline import handling_system_exit
from ..util.matplotlib import closing_figures
from ..util.tempfile import NamedTemporaryFile


@app.task(ignore_result=True, shared=False)
def annotate_fits(filecontents, versioned_filename, graceid, tags):
    """Perform annotations on a sky map.

    This function downloads a FITS file and then generates and uploads all
    derived images as well as an HTML dump of the FITS header.
    """
    multiorder_extension = '.multiorder.fits'
    flat_extension = '.fits'

    if multiorder_extension in versioned_filename:
        extension = multiorder_extension
        multiorder = True
    else:
        extension = flat_extension
        multiorder = False

    filebase, _, _ = versioned_filename.partition(extension)

    header_msg = (
        'FITS headers for <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    allsky_msg = (
        'Mollweide projection of <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    volume_msg = (
        'Volume rendering of <a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)
    flatten_msg = (
        'Flat-resolution fits file created from '
        '<a href="/api/superevents/{graceid}/files/'
        '{versioned_filename}">{versioned_filename}</a>').format(
            graceid=graceid, versioned_filename=versioned_filename)

    group(
        fits_header.s(versioned_filename)
        |
        gracedb.upload.s(
            filebase + '.html', graceid, header_msg, tags),

        plot_allsky.s()
        |
        gracedb.upload.s(
            filebase + '.png', graceid, allsky_msg, tags),

        annotate_fits_volume.s(
            filebase + '.volume.png', graceid, volume_msg, tags),

        *(
            [
                flatten.s(f'{filebase}.fits.gz')
                |
                gracedb.upload.s(
                    f'{filebase}.fits.gz', graceid, flatten_msg, tags)
            ] if multiorder else []
        )
    ).delay(filecontents)


def is_3d_fits_file(filecontents):
    """Determine if a FITS file has distance information."""
    with NamedTemporaryFile(content=filecontents) as fitsfile:
        return 'DISTNORM' in table.Table.read(fitsfile.name).colnames


@app.task(ignore_result=True, shared=False)
def annotate_fits_volume(filecontents, *args):
    """Perform annotations that are specific to 3D sky maps."""
    if is_3d_fits_file(filecontents):
        (
            plot_volume.s(filecontents)
            |
            gracedb.upload.s(*args)
        ).apply_async()


@app.task(shared=False)
def fits_header(filecontents, filename):
    """Dump FITS header to HTML."""
    template = env.get_template('fits_header.jinja2')
    with NamedTemporaryFile(content=filecontents) as fitsfile, \
            fits.open(fitsfile.name) as hdus:
        return template.render(filename=filename, hdus=hdus)


@app.task(shared=False)
@closing_figures()
def plot_allsky(filecontents, ra=None, dec=None):
    """Plot a Mollweide projection of a sky map using the command-line tool
    :doc:`ligo-skymap-plot <ligo.skymap:tool/ligo_skymap_plot>`.
    """
    # Explicitly use a non-interactive Matplotlib backend.
    plt.switch_backend('agg')

    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile, \
            handling_system_exit():
        if ra is not None and dec is not None:
            ligo_skymap_plot.main([fitsfile.name, '-o', pngfile.name,
                                   '--annotate', '--radec', str(ra), str(dec)])
        else:
            ligo_skymap_plot.main([fitsfile.name, '-o', pngfile.name,
                                   '--annotate', '--contour', '50', '90'])
        return pngfile.read()


@app.task(priority=1, queue='openmp', shared=False)
@closing_figures()
def plot_volume(filecontents):
    """Plot a 3D volume rendering of a sky map using the command-line tool
    :doc:`ligo-skymap-plot-volume <ligo.skymap:tool/ligo_skymap_plot_volume>`.
    """
    # Explicitly use a non-interactive Matplotlib backend.
    plt.switch_backend('agg')

    with NamedTemporaryFile(mode='rb', suffix='.png') as pngfile, \
            NamedTemporaryFile(content=filecontents) as fitsfile, \
            handling_system_exit():
        ligo_skymap_plot_volume.main([fitsfile.name, '-o',
                                      pngfile.name, '--annotate'])
        return pngfile.read()


@app.task(shared=False)
def flatten(filecontents, filename):
    """Convert a HEALPix FITS file from multi-resolution UNIQ indexing to the
    more common IMPLICIT indexing using the command-line tool
    :doc:`ligo-skymap-flatten <ligo.skymap:tool/ligo_skymap_flatten>`.
    """
    with NamedTemporaryFile(content=filecontents) as infile, \
            tempfile.TemporaryDirectory() as tmpdir, \
            handling_system_exit():
        outfilename = os.path.join(tmpdir, filename)
        ligo_skymap_flatten.main([infile.name, outfilename])
        return open(outfilename, 'rb').read()


@app.task(shared=False)
def unflatten(filecontents, filename):
    """Convert a HEALPix FITS file to multi-resolution UNIQ indexing from
    the more common IMPLICIT indexing using the command-line tool
    :doc:`ligo-skymap-unflatten <ligo.skymap:tool/ligo_skymap_unflatten>`.
    """
    with NamedTemporaryFile(content=filecontents) as infile, \
            tempfile.TemporaryDirectory() as tmpdir, \
            handling_system_exit():
        outfilename = os.path.join(tmpdir, filename)
        ligo_skymap_unflatten.main([infile.name, outfilename])
        return open(outfilename, 'rb').read()


@app.task(shared=False, queue='multiprocessing')
def skymap_from_samples(samplefilecontents):
    """Generate multi-resolution fits file from samples."""
    with NamedTemporaryFile(content=samplefilecontents) as samplefile, \
            tempfile.TemporaryDirectory() as tmpdir, \
            handling_system_exit():
        ligo_skymap_from_samples.main([
            '-j', '--seed', '150914', '--maxpts', '5000', '-o', tmpdir,
            samplefile.name])
        with open(os.path.join(tmpdir, 'skymap.fits'), 'rb') as f:
            return f.read()


def plot_bayes_factor(logb,
                      values=(1, 3, 5),
                      labels=('', 'strong', 'very strong'),
                      xlim=7, title=None, palette='RdYlBu',
                      var_label="B"):
    """Visualize a Bayes factor as a `bullet graph`_.

    Make a bar chart of a log Bayes factor as compared to a set of subjective
    threshold values. By default, use the thresholds from
    Kass & Raftery (1995).

    .. _`bullet graph`: https://en.wikipedia.org/wiki/Bullet_graph

    Parameters
    ----------
    logb : float
        The natural logarithm of the Bayes factor.
    values : list
        A list of floating point values for human-friendly confidence levels.
    labels : list
        A list of string labels for human-friendly confidence levels.
    xlim : float
        Limits of plot (`-xlim` to `+xlim`).
    title : str
        Title for plot.
    palette : str
        Color palette.
    var_label : str
        The variable symbol used in plotting

    Returns
    -------
    fig : Matplotlib figure
    ax : Matplotlib axes

    Example
    -------
    .. plot::
        :alt: Example Bayes factor plot

        from gwcelery.tasks.skymaps import plot_bayes_factor
        plot_bayes_factor(6.3, title='GWCelery is awesome')

    """
    with plt.style.context('seaborn-notebook'):
        fig, ax = plt.subplots(figsize=(6, 1.7), tight_layout=True)
        ax.set_xlim(-xlim, xlim)
        ax.set_ylim(-0.5, 0.5)
        ax.set_yticks([])
        ax.set_title(title)
        ax.set_ylabel(r'$\ln\,{}$'.format(var_label), rotation=0,
                      rotation_mode='anchor',
                      ha='right', va='center')

        # Add human-friendly labels
        ticks = (*(-x for x in reversed(values)), 0, *values)
        ticklabels = (
            *(f'{s}\nevidence\nagainst'.strip() for s in reversed(labels)), '',
            *(f'{s}\nevidence\nfor'.strip() for s in labels))
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticklabels)
        plt.setp(ax.get_xticklines(), visible=False)
        plt.setp(ax.get_xticklabels()[:len(ticks) // 2], ha='right')
        plt.setp(ax.get_xticklabels()[len(ticks) // 2:], ha='left')

        # Plot colored bands for confidence thresholds
        fmt = plt.FuncFormatter(lambda x, _: f'{x:+g}'.replace('+0', '0'))
        ax2 = ax.twiny()
        ax2.set_xlim(*ax.get_xlim())
        ax2.set_xticks(ticks)
        ax2.xaxis.set_major_formatter(fmt)
        levels = (-xlim, *ticks, xlim)
        colors = plt.get_cmap(palette)(np.arange(1, len(levels)) / len(levels))
        ax.barh(0, np.diff(levels), 1, levels[:-1],
                linewidth=plt.rcParams['xtick.major.width'],
                color=colors, edgecolor='white')

        # Plot bar for log Bayes factor value
        ax.barh(0, logb, 0.5, color='black',
                linewidth=plt.rcParams['xtick.major.width'],
                edgecolor='white')

        for ax_ in fig.axes:
            ax_.grid(False)
            for spine in ax_.spines.values():
                spine.set_visible(False)

    return fig, ax


@app.task(shared=False)
@closing_figures()
def plot_coherence(filecontents):
    """IGWN alert handler to plot the coherence Bayes factor.

    Parameters
    ----------
    contents : str, bytes
        The contents of the FITS file.

    Returns
    -------
    png : bytes
        The contents of a PNG file.

    Notes
    -----
    Under the hood, this just calls :meth:`plot_bayes_factor`.

    """
    # Explicitly use a non-interactive Matplotlib backend.
    plt.switch_backend('agg')

    with NamedTemporaryFile(content=filecontents) as fitsfile:
        header = fits.getheader(fitsfile, 1)
    try:
        logb = header['LOGBCI']
    except KeyError:
        raise Ignore('FITS file does not have a LOGBCI field')

    objid = header['OBJECT']
    logb_string = np.format_float_positional(logb, 1, trim='0', sign=True)
    fig, _ = plot_bayes_factor(
        logb, title=rf'Coherence of {objid} $[\ln\,B = {logb_string}]$')
    outfile = io.BytesIO()
    fig.savefig(outfile, format='png')
    return outfile.getvalue()


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    shared=False)
def handle_plot_coherence(alert):
    """IGWN alert handler to plot and upload a visualization of the coherence
    Bayes factor.

    Notes
    -----
    Under the hood, this just calls :meth:`plot_coherence`.

    """
    if alert['alert_type'] != 'log':
        return  # not for us
    if not alert['data']['filename'].endswith('.fits') and \
            (alert['data']['filename'] not in
             {external_skymaps.COMBINED_SKYMAP_FILENAME_MULTIORDER,
              external_skymaps.COMBINED_SKYMAP_FILENAME_FLAT}):
        return  # not for us

    graceid = alert['uid']
    f = alert['data']['filename']
    v = alert['data']['file_version']
    fv = '{},{}'.format(f, v)

    (
        gracedb.download.s(fv, graceid)
        |
        plot_coherence.s()
        |
        gracedb.upload.s(
            f.replace('.fits', '.coherence.png'), graceid,
            message=(
                f'Bayes factor for coherence vs. incoherence of '
                f'<a href="/api/superevents/{graceid}/files/{fv}">'
                f'{fv}</a>'),
            tags=['sky_loc']
        )
    ).delay()



"""Application configuration for ``minikube`` local installation."""

from . import *  # noqa: F401, F403

expose_to_public = True
"""Set to True if events meeting the public alert threshold really should be
exposed to the public."""

igwn_alert_server = 'kafka://hopskotch-server'
"""IGWN alert server: None == DEFAULT_SERVER"""

igwn_alert_noauth = True
"""IGWN alert server no-authetication"""

igwn_alert_group = 'default'
"""IGWN alert group."""

gracedb_host = 'gracedb.default.svc.cluster.local'
"""GraceDB host."""

early_warning_alert_far_threshold = float('inf')
"""False alarm rate threshold for early warning alerts."""

mock_events_simulate_multiple_uploads = False
"""If True, then upload each mock event several times in rapid succession with
random jitter in order to simulate multiple pipeline uploads."""

kafka_alert_config = {
    'scimma': {'url': 'kafka://hopskotch-server/igwn.gwalert-minikube',
               'suffix': 'avro', 'skymap_encoder': lambda _: _}
}
"""Kafka broker configuration details"""



import json

import pytest

from ..tasks import p_astro


@pytest.mark.parametrize(
    'snr,far,mass1,mass2,pipeline,instruments',
    ([30., 1e-17, 1.4, 1.4, 'mbta', {'H1', 'L1', 'V1'}],
     [100, 3e-10, 400, 40, 'mbta', {'H1', 'L1'}],
     [10, 3e-10, 20, 30, 'pycbc', {'L1', 'V1'}]))
def test_compute_p_astro(snr, far, mass1, mass2,
                         pipeline, instruments):
    """Check if p_astros sum up to unity"""
    p_astros = json.loads(
        p_astro.compute_p_astro(snr, far, mass1, mass2,
                                pipeline, instruments))
    assert (
        set(p_astros) - set({'BNS', 'Terrestrial', 'NSBH', 'BBH'})
    ) == set()
    assert pytest.approx(sum(p_astros.values())) == 1.


@pytest.mark.parametrize(
    'far,pipeline,instruments,snr_thresh,val',
    ([1e-25, 'mbta', {'H1', 'L1', 'V1'}, 12, 1],
     [1e-8, 'mbta', {'L1', 'V1'}, 33, 0.08],
     [6e-10, 'mbta', {'H1', 'V1'}, 10, 0.99],
     [7.6e-59, 'spiir', {'H1', 'L1', 'V1'}, 33, 1],
     [1e-10, 'pycbc', {'H1', 'L1'}, 10, 1],
     [1e-10, 'pycbc', {'H1', 'L1', 'V1'}, 10, 1]))
def test_compute_p_astro_bns(far, pipeline, instruments,
                             snr_thresh, val):
    """Test p_astro values using CBC catalog paper
    values for GW170817, for various mock-FARs, to test
    handling of this very loud event for MBTA, PyCBC
    and spiir.
    """
    # values based on G322759
    snr = 33.
    mass1 = 1.77
    mass2 = 1.07

    p_astros = json.loads(
        p_astro.compute_p_astro(snr, far, mass1, mass2,
                                pipeline, instruments))

    assert pytest.approx(p_astros['BNS'], abs=1e-1) == val



from importlib import resources
from unittest.mock import patch

import pytest

from ..tasks.gcn import validate
from ..util import read_json
from . import data


@pytest.fixture
def fake_gcn(monkeypatch):

    def mock_download(filename, graceid):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        return resources.read_binary(data, filename)

    def mock_get_log(graceid):
        assert graceid == 'G298048'
        return read_json(data, 'G298048_log.json')

    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.download', mock_download)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_log', mock_get_log)

    # Get the VOEvent.
    yield resources.read_binary(data, 'G298048-1-Initial.xml')


@patch('gwcelery.tasks.gracedb.create_tag.run')
@patch('gwcelery.tasks.gracedb.upload.run')
def test_validate_voevent(mock_upload, mock_create_tag, fake_gcn):
    """Test that the fake GCN notice matches what we actually sent."""
    validate(fake_gcn)
    mock_create_tag.assert_called_once_with(
        'G298048-1-Initial.xml', 'gcn_received', 'G298048')
    mock_upload.assert_not_called()


@patch('gwcelery.tasks.gracedb.create_tag.run')
@patch('gwcelery.tasks.gracedb.upload.run')
def test_validate_voevent_mismatched_param(
        mock_upload, mock_create_tag, fake_gcn):
    """Test that we correctly detect mismatched parameter values in GCNs."""
    with pytest.raises(
            ValueError,
            match="^VOEvent received from GCN differs from what we sent"):
        validate(fake_gcn + b'\n')
    mock_create_tag.assert_not_called()
    mock_upload.assert_called_once()



"""Communication with GraceDB."""
from requests.exceptions import ConnectionError, HTTPError
import functools
import re

from celery.utils.log import get_task_logger
import gracedb_sdk

from ..import app
from ..util import PromiseProxy

client = PromiseProxy(gracedb_sdk.Client,
                      ('https://' + app.conf.gracedb_host + '/api/',),
                      {'fail_if_noauth': True, 'cert_reload': True})

log = get_task_logger(__name__)


class RetryableHTTPError(HTTPError):
    """Exception class for server-side HTTP errors that we should retry."""


def catch_retryable_http_errors(f):
    """Decorator to capture server-side errors that we should retry.

    We retry HTTP status 502 (Bad Gateway), 503 (Service Unavailable), and
    504 (Gateway Timeout).
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except HTTPError as e:
            if e.response.status_code in {408, 429, 502, 503, 504}:
                raise RetryableHTTPError(
                    *e.args, request=e.request, response=e.response)
            else:
                raise

    return wrapper


def task(*args, **kwargs):
    return app.task(*args, **kwargs,
                    autoretry_for=(
                        ConnectionError, RetryableHTTPError, TimeoutError),
                    default_retry_delay=20.0, retry_backoff=True,
                    retry_kwargs=dict(max_retries=10))


versioned_filename_regex = re.compile(
    r'^(?P<filename>.*?)(?:,(?P<file_version>\d+))?$')


def _parse_versioned_filename(versioned_filename):
    match = versioned_filename_regex.fullmatch(versioned_filename)
    filename = match['filename']
    file_version = match['file_version']
    if file_version is not None:
        file_version = int(file_version)
    return filename, file_version


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def add_pipeline_preferred_event(superevent_id, event_id):
    client.superevents[superevent_id].pipeline_preferred_events.add(event_id)


@task(shared=False)
@catch_retryable_http_errors
def create_event(filecontents, search, pipeline, group, labels=()):
    """Create an event in GraceDB."""
    response = client.events.create(group=group, pipeline=pipeline,
                                    filename='initial.data', search=search,
                                    filecontents=filecontents, labels=labels)
    return response


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def create_label(label, graceid):
    """Create a label in GraceDB."""
    try:
        client.events[graceid].labels.create(label)
    except HTTPError as e:
        # If we got a 400 error because no change was made, then ignore
        # the exception and return successfully to preserve idempotency.
        messages = {
            b'"The \'ADVREQ\' label cannot be applied to request a signoff '
            b'because a related signoff already exists."',

            b'"The fields superevent, name must make a unique set."'
        }
        if e.response.content not in messages:
            raise


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def remove_label(label, graceid):
    """Remove a label in GraceDB."""
    try:
        client.events[graceid].labels.delete(label)
    except HTTPError as e:
        # If the label did not exist, then GraceDB will return a 404 error.
        # Don't treat this as a failure because we got what we wanted: for the
        # label to be removed.
        if e.response.status_code != 404:
            raise


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def create_signoff(status, comment, signoff_type, graceid):
    """Create a signoff in GraceDB."""
    try:
        client.superevents[graceid].signoff(signoff_type, status, comment)
    except HTTPError as e:
        # If we got a 400 error because the signoff was already applied,
        # then ignore the exception and return successfully to preserve
        # idempotency.
        message = b'The fields superevent, instrument must make a unique set'
        if message not in e.response.content:
            raise


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def create_tag(filename, tag, graceid):
    """Create a tag in GraceDB."""
    filename, file_version = _parse_versioned_filename(filename)
    log = get_log(graceid)
    if file_version is None:
        *_, entry = (e for e in log if e['filename'] == filename)
    else:
        *_, entry = (e for e in log if e['filename'] == filename
                     and e['file_version'] == file_version)
    log_number = entry['N']
    try:
        client.events[graceid].logs[log_number].tags.create(tag)
    except HTTPError as e:
        # If we got a 400 error because no change was made, then ignore
        # the exception and return successfully to preserve idempotency.
        message = b'"Tag is already applied to this log message"'
        if e.response.content != message:
            raise


@task(shared=False)
@catch_retryable_http_errors
def create_voevent(graceid, voevent_type, **kwargs):
    """Create a VOEvent.

    Returns
    -------
    str
        The filename of the new VOEvent.

    """
    response = client.events[graceid].voevents.create(
        voevent_type=voevent_type, **kwargs)
    return response['filename']


@task(shared=False)
@catch_retryable_http_errors
def download(filename, graceid):
    """Download a file from GraceDB."""
    with client.events[graceid].files[filename].get() as f:
        return f.read()


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def expose(graceid):
    """Expose an event to the public.

    Notes
    -----
    If :obj:`~gwcelery.conf.expose_to_public` is False, then this because a
    no-op.

    """
    if app.conf['expose_to_public']:
        client.superevents[graceid].expose()


@task(shared=False)
@catch_retryable_http_errors
def get_events(query, **kwargs):
    """Get events from GraceDB."""
    return list(client.events.search(query=query, **kwargs))


@task(shared=False)
@catch_retryable_http_errors
def get_event(graceid):
    """Retrieve an event from GraceDB."""
    return client.events[graceid].get()


@task(shared=False)
@catch_retryable_http_errors
def get_group(graceid):
    """Retrieve the search field of an event from GraceDB."""
    return client.events[graceid].get()['group']


@task(shared=False)
@catch_retryable_http_errors
def get_search(graceid):
    """Retrieve the search field of an event from GraceDB."""
    return client.events[graceid].get()['search']


@task(shared=False)
@catch_retryable_http_errors
def get_labels(graceid):
    """Get all labels for an event in GraceDB."""
    return {row['name'] for row in client.events[graceid].labels.get()}


@task(shared=False)
@catch_retryable_http_errors
def get_log(graceid):
    """Get all log messages for an event in GraceDB."""
    return client.events[graceid].logs.get()


@task(shared=False)
@catch_retryable_http_errors
def get_superevent(graceid):
    """Retrieve a superevent from GraceDB."""
    return client.superevents[graceid].get()


@task(shared=False)
@catch_retryable_http_errors
def replace_event(graceid, payload):
    """Get an event from GraceDB."""
    return client.events.update(graceid, filecontents=payload)


@task(shared=False)
@catch_retryable_http_errors
def upload(filecontents, filename, graceid, message, tags=()):
    """Upload a file to GraceDB."""
    result = client.events[graceid].logs.create(
        comment=message, filename=filename,
        filecontents=filecontents, tags=tags)
    return '{},{}'.format(result['filename'], result['file_version'])


@app.task(shared=False)
@catch_retryable_http_errors
def get_superevents(query, **kwargs):
    """List matching superevents in gracedb.

    Parameters
    ----------
    *args
        arguments passed to :meth:`GraceDb.superevents`
    **kwargs
        keyword arguments passed to :meth:`GraceDb.superevents`

    Returns
    -------
    superevents : list
        The list of the superevents.

    """
    return list(client.superevents.search(query=query, **kwargs))


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def remove_pipeline_preferred_event(superevent_id, event_id):
    client.superevents[
        superevent_id].pipeline_preferred_events.remove(event_id)


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def update_superevent(superevent_id, t_start=None,
                      t_end=None, t_0=None, preferred_event=None,
                      em_type=None, time_coinc_far=None,
                      space_coinc_far=None):
    """
    Update superevent information. Wrapper around
    :meth:`updateSuperevent`

    Parameters
    ----------
    superevent_id : str
        superevent uid
    t_start : float
        start of superevent time window, unchanged if None
    t_end : float
        end of superevent time window, unchanged if None
    t_0 : float
        superevent t_0, unchanged if None
    preferred_event : str
        uid of the preferred event, unchanged if None

    """
    try:
        client.superevents.update(
            superevent_id, t_start=t_start, t_end=t_end, t_0=t_0,
            preferred_event=preferred_event, em_type=em_type,
            time_coinc_far=time_coinc_far, space_coinc_far=space_coinc_far)
    except HTTPError as e:
        # If we got a 400 error because no change was made, then ignore
        # the exception and return successfully to preserve idempotency.
        error_msg = b'"Request would not modify the superevent"'
        if not (e.response.status_code == 400
                and e.response.content == error_msg):
            raise


@task(shared=False)
@catch_retryable_http_errors
def create_superevent(graceid, t0, t_start, t_end):
    """Create new superevent in GraceDB with `graceid`

    Parameters
    ----------
    graceid : str
        graceid with which superevent is created.
    t0 : float
        ``t_0`` parameter of superevent
    t_start : float
        ``t_start`` parameter of superevent
    t_end : float
        ``t_end`` parameter of superevent

    """
    try:
        response = client.superevents.create(
            t_start=t_start, t_0=t0, t_end=t_end, preferred_event=graceid)
        return response['superevent_id']
    except HTTPError as e:
        error_msg = b'is already assigned to a Superevent'
        if not (e.response.status_code == 400
                and error_msg in e.response.content):
            raise


@task(ignore_result=True, shared=False)
@catch_retryable_http_errors
def add_event_to_superevent(superevent_id, graceid):
    """Add an event to a superevent in GraceDB."""
    try:
        client.superevents[superevent_id].add(graceid)
    except HTTPError as e:
        error_msg = b'is already assigned to a Superevent'
        if not (e.response.status_code == 400
                and error_msg in e.response.content):
            raise



"""A Nagios plugin for monitoring GWCelery.

See https://nagios-plugins.org/doc/guidelines.html.
"""
from enum import IntEnum
from sys import exit
from traceback import format_exc, format_exception

import click
import kombu.exceptions

# Make sure that all tasks are registered
from .. import tasks  # noqa: F401


class NagiosPluginStatus(IntEnum):
    """Nagios plugin status codes."""

    OK = 0
    WARNING = 1
    CRITICAL = 2
    UNKNOWN = 3


class NagiosCriticalError(Exception):
    """An exception that maps to a Nagios status of `CRITICAL`."""


def get_active_queues(inspector):
    return {queue['name']
            for queues in (inspector.active_queues() or {}).values()
            for queue in queues}


def get_active_igwn_alert_topics(inspector):
    return {node for stat in inspector.stats().values()
            for node in stat.get('igwn-alert-topics', ())}


def get_expected_queues(app):
    # Get the queues for all registered tasks.
    result = {getattr(task, 'queue', None) for task in app.tasks.values()}
    # We use 'celery' for all tasks that do not explicitly specify a queue.
    result -= {None}
    result |= {'celery'}
    # Done.
    return result


def get_expected_igwn_alert_topics(app):
    return app.conf['igwn_alert_topics']


def get_active_voevent_peers(inspector):
    stats = inspector.stats()
    broker_peers, receiver_peers = (
        {peer for stat in stats.values() for peer in stat.get(key, ())}
        for key in ['voevent-broker-peers', 'voevent-receiver-peers'])
    return broker_peers, receiver_peers


def get_expected_kafka_bootstep_urls(inspector):
    stats = inspector.stats()
    expected_kafka_urls = \
        {peer for stat in stats.values() for peer in
         stat.get('kafka_topic_up', {})}
    return expected_kafka_urls


def get_active_kafka_bootstep_urls(inspector):
    stats = inspector.stats()
    active_kafka_urls = \
        {kafka_url for stat in stats.values() for kafka_url, active_flag in
         stat.get('kafka_topic_up', {}).items() if active_flag}
    return active_kafka_urls


def get_undelivered_message_urls(inspector):
    stats = inspector.stats()
    undelievered_messages = \
        {kafka_url for stat in stats.values() for kafka_url, active_flag in
         stat.get('kafka_delivery_failures', {}).items() if active_flag}
    return undelievered_messages


def get_active_kafka_consumer_bootstep_names(inspector):
    stats = inspector.stats()
    active_kafka_consumer_urls = {consumer for stat in stats.values() for
                                  consumer in stat.get(
                                      'active_kafka_consumers', ()
                                  )}
    return active_kafka_consumer_urls


def get_expected_kafka_consumer_bootstep_names(app):
    return {name for name in app.conf['kafka_consumer_config'].keys()}


def get_celery_queue_length(app):
    return app.backend.client.llen("celery")


def check_status(app):
    connection = app.connection()
    try:
        connection.ensure_connection(max_retries=1)
    except kombu.exceptions.OperationalError as e:
        raise NagiosCriticalError('No connection to broker') from e

    inspector = app.control.inspect()

    active = get_active_queues(inspector)
    expected = get_expected_queues(app)
    missing = expected - active
    if missing:
        raise NagiosCriticalError('Not all expected queues are active') from \
              AssertionError('Missing queues: ' + ', '.join(missing))

    active = get_active_igwn_alert_topics(inspector)
    expected = get_expected_igwn_alert_topics(app)
    missing = expected - active
    extra = active - expected
    if missing:
        raise NagiosCriticalError('Not all IGWN alert topics are subscribed') \
            from AssertionError('Missing topics: ' + ', '.join(missing))
    if extra:
        raise NagiosCriticalError(
            'Too many IGWN alert topics are subscribed') from AssertionError(
                    'Extra topics: ' + ', '.join(extra))

    broker_peers, receiver_peers = get_active_voevent_peers(inspector)
    if app.conf['voevent_broadcaster_whitelist'] and not broker_peers:
        raise NagiosCriticalError(
            'The VOEvent broker has no active connections') \
                from AssertionError('voevent_broadcaster_whitelist: {}'.format(
                    app.conf['voevent_broadcaster_whitelist']))
    if app.conf['voevent_receiver_address'] and not receiver_peers:
        raise NagiosCriticalError(
            'The VOEvent receiver has no active connections') \
                from AssertionError('voevent_receiver_address: {}'.format(
                    app.conf['voevent_receiver_address']))

    active = get_active_kafka_bootstep_urls(inspector)
    expected = get_expected_kafka_bootstep_urls(inspector)
    missing = expected - active
    if missing:
        raise NagiosCriticalError('Not all Kafka bootstep URLs are active') \
            from AssertionError('Missing urls: ' + ', '.join(missing))

    undelivered_messages = get_undelivered_message_urls(inspector)
    if undelivered_messages:
        raise NagiosCriticalError(
            'Not all Kafka messages have been succesfully delivered'
        ) from AssertionError(
                'URLs with undelivered messages: ' + ', '.join(missing)
        )

    celery_queue_length = get_celery_queue_length(app)
    if celery_queue_length > 50:
        raise NagiosCriticalError(
            'Tasks are piled up in Celery queue') from AssertionError(
                'Length of celery queue is {}'.format(celery_queue_length))

    active = get_active_kafka_consumer_bootstep_names(inspector)
    expected = get_expected_kafka_consumer_bootstep_names(app)
    missing = expected - active
    if missing:
        raise NagiosCriticalError('Not all Kafka consumer bootstep topics are '
                                  'active') \
            from AssertionError('Missing urls: ' + ', '.join(missing))


@click.command(help=__doc__)
@click.pass_context
def nagios(ctx):
    try:
        check_status(ctx.obj.app)
    except NagiosCriticalError as e:
        status = NagiosPluginStatus.CRITICAL
        output, = e.args
        e = e.__cause__
        detail = ''.join(format_exception(type(e), e, e.__traceback__))
    except:  # noqa: E722
        status = NagiosPluginStatus.UNKNOWN
        output = 'Unexpected error'
        detail = format_exc()
    else:
        status = NagiosPluginStatus.OK
        output = 'Running normally'
        detail = None
    print('{}: {}'.format(status.name, output))
    if detail:
        print(detail)
    exit(status)
